from pyplan_core.classes.dynamics.BaseDynamic import BaseDynamic
from pyplan_core.cubepy.Helpers import Helpers
import numpy as np
from pyplan_core import cubepy 
import datetime as dt
import re


class CubepyDynamic(BaseDynamic):

    def circularEval(self, node, params):
        """
        Used for execute nodes with circular reference (dynamic)
        """

        dynamicVars = params["dynamicVars"]
        dynamicIndex = params["dynamicIndex"]
        nodesInCyclic = params["nodesInCyclic"]
        initialValues = params["initialValues"]
        shift = params["shift"]

        startTime = dt.datetime.now()

        evaluate = node.model.evaluate

        # create nodes array
        cyclicNodes = []

        try:
            node.model.inCyclicEvaluate = True
            for nodeId in nodesInCyclic:
                _nodeObj = node.model.getNode(nodeId)
                cyclicNodes.append({
                    "node": _nodeObj,
                    "initialize": self.generateInitDef(node, _nodeObj.bypassCircularEvaluator().result, dynamicIndex),
                    "loopDefinition": self.generateLoopDef(node, _nodeObj.definition, nodesInCyclic)
                })
        except Exception as e:
            raise e
        finally:
            node.model.inCyclicEvaluate = False

        cyclicDic = {}

        # initialice var dictionary
        for _node in cyclicNodes:
            _id = _node["node"].identifier
            cyclicDic[_id] = evaluate(_node["initialize"])
            if not initialValues is None and _id in initialValues:
                cyclicDic[_id] = cyclicDic[_id] + evaluate(initialValues[_id])

        # initialice vars in t-1
        for _var in dynamicVars:
            _key = "__" + _var + "_t"
            cyclicDic[_key] = cyclicDic[_var].sum(dynamicIndex) * 0

        # loop over index
        cyclicParams = None
        # for item in dynamicIndex:

        theRange = range(0, len(dynamicIndex))
        initialCount = shift*-1
        reverseMode = False
        if shift > 0:
            theRange = range(len(dynamicIndex)-1, -1, -1)
            initialCount = len(dynamicIndex)-1-shift
            reverseMode = True

        for nn in theRange:
            item = dynamicIndex[nn]

            # load params
            cyclicParams = {
                "item": item,
                "cyclicDic": cyclicDic
            }

            __initialValues = None
            # loop over variables
            for _node in cyclicNodes:

                _id = _node["node"].identifier
                node.model.currentProcessingNode(_id)
                cyclicParams["cyclicDic"] = cyclicDic

                # execute vars
                if (_id in initialValues) and ((nn < initialCount and (not reverseMode)) or (nn > initialCount and reverseMode)):
                    # use initial values
                    __resultNode = evaluate(
                        _node["loopDefinition"], cyclicParams)

                    __initialValues = evaluate(initialValues[_id])
                    if isinstance(__initialValues, cubepy.Cube):
                        __finalNode = __initialValues.tryFilter(
                            item).squeeze() + __resultNode.tryFilter(item).squeeze()
                    else:
                        __finalNode = __resultNode.tryFilter(
                            item).squeeze() + __initialValues

                    cyclicDic[_id].set_data(item, __finalNode.values)
                else:
                    # dont use use initial values
                    __resultNode = evaluate(
                        _node["loopDefinition"], cyclicParams)
                    cyclicDic[_id].set_data(
                        item, __resultNode.tryFilter(item).squeeze().values)

            # set dynamicVar
            # if ( (not reverseMode) and ((nn+1)<initialCount or ((((nn+1) % abs(shift)) != 0) ))) or (reverseMode and ((nn-1)>initialCount or (((len(dynamicIndex)-1-(nn-1)) % abs(shift)) != 0))):
            if (not reverseMode and (nn+1) < initialCount) or (reverseMode and (nn-1) > initialCount):
                # do not update t- vars
                pass
            else:
                for _var in dynamicVars:
                    _key = "__" + _var + "_t"
                    #cyclicDic[_key] = cyclicDic[_var].tryFilter(item).squeeze()
                    if reverseMode:
                        cyclicDic[_key] = cyclicDic[_var].tryFilter(
                            dynamicIndex[nn+shift-1]).squeeze()
                    else:
                        cyclicDic[_key] = cyclicDic[_var].tryFilter(
                            dynamicIndex[nn-initialCount+1]).squeeze()

        endTime = dt.datetime.now()

        # set result
        for _node in cyclicNodes:
            _id = _node["node"].identifier
            _node["node"]._result = cyclicDic[_id]
            _node["node"].__resultMemory = Helpers.getResultSize(
                _node["node"]._result)
            _node["node"]._isCalc = True
            _node["node"].lastEvaluationTime = (
                endTime - startTime).total_seconds()
            _node["node"].evaluationVersion = node.model.evaluationVersion
            _node["node"]._bypassCircularEvaluator = False

        evaluate = None
        model = None
        cyclicDic = None
        cyclicParams = None

        return "ok"

    def generateLoopDef(self, node, nodeDefinition, cyclicVariables):
        """
        Return definition used for circular evaluator
        """
        _def = self.clearCircularDependency(
            nodeDefinition, "cyclicDic['__##node##_t']")
        finalDef = _def
        tmpCode = compile(_def, '<string>', 'exec')
        names = node.parseNames(tmpCode)
        rx = r"('[^'\\]*(?:\\.[^'\\]*)*'|\"[^\"\\]*(?:\\.[^\"\\]*)*\")|\b{0}\b"
        for nodeId in names:
            if node._model.existNode(node._model.clearId(nodeId)):
                if nodeId in cyclicVariables:
                    finalDef = re.sub(rx.format(nodeId), lambda m:
                                      (
                        m.group(1)
                        if m.group(1)
                        else
                        (
                            "cyclicDic['"+nodeId+"']"
                            if (m.endpos > m.regs[0][1]+5) and ((m.string[m.regs[0][1]:m.regs[0][1]+5] == '.node') or (m.string[m.regs[0][1]:m.regs[0][1]+8] == '.timeit('))
                            else
                            (nodeId
                             if (m.string[m.regs[0][0]-1:m.regs[0][0]+len(nodeId)] == ('.'+nodeId))
                             else "cyclicDic['"+nodeId+"'].tryFilter(item).squeeze() "
                             )
                        )
                    ), finalDef, 0, re.IGNORECASE)
        return finalDef

    def generateInitDef(self, node, nodeCube, dynamicIndex):
        """
        Return definition used for initialice vars in circular evaluator
        """
        if isinstance(nodeCube, cubepy.Cube):
            _list = nodeCube.dims[:]
            if not dynamicIndex.name in _list:
                _list.append(dynamicIndex.name)
            _dims = ','.join(_list)
            _def = "result = cp.cube([" + _dims + "])"
            return _def
        else:
            return "result = cp.cube([" + dynamicIndex.name + "])"

    def generateCircularParameters(self, node, nodeList):
        """
        Generate paremters for call to circularEval
        """
        dynamicVars = []
        dynamicIndex = None
        nodesInCyclic = []  # nodeList TODO: Determinar orden de nodos
        initialValues = {}
        shift = -1

        for _nodeId in nodeList:
            if node.model.existNode(_nodeId):
                _def = node.model.getNode(_nodeId).definition
                if "cp.dynamic" in _def:
                    _startPos = _def.find("cp.dynamic(") + 11
                    _endPos = _def.find(")", _startPos)

                    # dynamicVars = _def[_startPos:_endPos] # cc,time,-1
                    _arr = str(_def[_startPos:_endPos]).split(",")
                    _vart_1 = _arr[0].strip()
                    if not _vart_1 in dynamicVars:
                        dynamicVars.append(_vart_1)
                    dynamicIndex = node.model.getNode(_arr[1].strip()).result
                    shift = int(_arr[2])
                    if len(_arr) > 3:
                        initialValues[_nodeId] = "result = " + _arr[3].strip()

        #nodesInCyclic =  sort

        _graph = {}

        for _nodeId in nodeList:
            if node.model.existNode(_nodeId):
                _graph[_nodeId] = []
                for _outputId in node.model.getNode(_nodeId).ioEngine.outputs:
                    if _outputId in nodeList:
                        if 'cp.dynamic('+_nodeId+',' not in str(node.model.getNode(_outputId).definition).replace(" ", ""):
                            _graph[_nodeId].append(_outputId)

        nodesInCyclic = Helpers.dfs_topsort(_graph)

        return {
            "dynamicVars": dynamicVars,
            "dynamicIndex": dynamicIndex,
            "nodesInCyclic": nodesInCyclic,
            "initialValues": initialValues,
            "shift": shift
        }

    def clearCircularDependency(self, stringDef, replaceWith="0"):
        """ Replaces cp.dynamic(x,y,z) for the desired replaceWith param
        """
        response = stringDef
        initialIndex = -1
        startIndex = -1
        finalIndex = -1
        toReplace = ''
        initialIndex = stringDef.find('cp.dynamic(')
        if initialIndex != -1:
            startIndex = initialIndex
            initialIndex = initialIndex + len('cp.dynamic(')
            if len(stringDef) > initialIndex:
                finalIndex = stringDef[initialIndex+1:].find(')')

        if initialIndex != -1 and finalIndex != -1:
            toReplace = stringDef[startIndex:initialIndex + finalIndex + 2]
            if "cyclicDic" in replaceWith:
                nodeInT1 = toReplace.split(",")[0]
                nodeInT1 = nodeInT1[(nodeInT1.find("(")+1):]
                replaceWith = replaceWith.replace("##node##", nodeInT1.strip())
            response = stringDef.replace(toReplace, replaceWith)

        return response



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import json


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "DataarrayFromPandas"

    def getDataframeSchema(self, model, params):
        res = dict()
        res["columns"] = self.getColumnList(model, params)
        res["model_indexes"] = [{"title": node.title, "id": node.identifier}
                                for node in model.findNodes('nodeClass', "index")]
        res["model_indexes"].sort(key=lambda x: x["id"])
        return res

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]

        if model.existNode(nodeId):
            base_node = model.getNode(nodeId)
            index_list = []
            domain_dic = dict()
            agg_dic = dict()
            new_def = "# Preparing the dataframe to be converted to dataarray\n"
            value_colum = ""
            previous_step = nodeId

            for index_dic in params["indexes"]:
                if not index_dic["model_index"]:
                    column_name = index_dic["field"]
                    new_index_id = model._removeDiacritics(column_name)
                    if model.existNode(new_index_id):
                        new_index_id = model.getNextIdentifier(new_index_id)

                    node_position = model.getAPlace(base_node.moduleId, int(
                        base_node.x) + 150, int(base_node.y) - 70, 96, 48)
                    new_index_node = model.createNode(
                        new_index_id, nodeClass="index", moduleId=base_node.moduleId, x=node_position["x"], y=node_position["y"])
                    new_index_node.title = column_name
                    new_index_node.definition = f"result = pd.Index({nodeId}.reset_index()['{column_name}'].unique().tolist())"
                    index_dic["model_index"] = new_index_node.identifier
                index_list.append(index_dic["field"])
                domain_dic[index_dic["field"]
                           ] = f"_#_{index_dic['model_index']}_#_"

            agg_item = params["agg"]
            column_name = agg_item["field"]
            calc = agg_item["calc"]
            value_colum = column_name
            new_node_title = f"{calc.capitalize()} of {column_name}"
            agg_dic[column_name] = calc

            # create dataarray node
            new_node_id = model._removeDiacritics(new_node_title)
            if model.existNode(new_node_id):
                new_node_id = model.getNextIdentifier(new_node_id)

            node_position = model.getAPlace(base_node.moduleId, int(
                base_node.x) + 150, int(base_node.y), 96, 48)
            new_node = model.createNode(new_node_id, nodeClass="variable",
                                        moduleId=base_node.moduleId, x=node_position["x"], y=node_position["y"])
            new_node.title = new_node_title
            domain_dic_str = json.dumps(domain_dic).replace(
                "\"_#_", "").replace("_#_\"", "")

            new_def += f"_df = {previous_step}.reset_index().groupby({json.dumps(index_list)}, as_index=False).agg({agg_dic})\n"
            new_def += f"result = pp.dataarray_from_pandas(_df, {domain_dic_str}, '{value_colum}')"
            new_node.definition = new_def

            return new_node.identifier

        return ""



#functions for use as:  import pyplan_core.cubepy.factory as cp

import numpy as np
from pyplan_core import cubepy 
import numbers
from sys import platform
from pyplan_core.cubepy.cube import kindToString, apply_op
import os
import csv

random = -123798
byVal = 1
byPos = 2
exact = 1
start_with = 2
end_with = 3
contain = 4




def cube(axes, values=None, broadcast=True,dtype=None):
    """Create a cube  object. 
    axes: list of axis of the cube
    values: optional, list of values of the cube. Can be other cubes for build a report.
    Ex.
        cp.cube([time])
        cp.cube([time,product])
        cp.cube([time,product],[10,20,30,40,50,60,70,80])
        cp.cube([time,product],cp.random)
        
        cp.cube([index_reports],[report_1,report_2])
    """
    if values is None:
        if not dtype is None:
            if dtype is str:
                return cubepy.Cube.full(axes,'', dtype='U25')
            elif kindToString(np.dtype(dtype).kind)=="string":
                return cubepy.Cube.full(axes,'', dtype=dtype)
        return cubepy.Cube.zeros(axes)
    else:
        if isinstance(values,list) or isinstance(values,np.ndarray) :
            if len(values)>0:
                if isinstance(values[0],cubepy.Cube):
                    #use stack
                    if isinstance(axes,list):
                        axes = axes[0]
                    return cubepy.stack(values,axes,broadcast)
            return cubepy.Cube(axes,values,fillValues=True, dtype=dtype)
        elif isinstance(values,numbers.Number) and values==random:
            theSize = [len(x) for x in axes]
            return cube(axes, np.random.randint(100, size=theSize), dtype=dtype)
        else:
            return cubepy.Cube.full(axes,values, dtype=dtype)


def index(name,values):
    """Create a index object.
    name: name for the index
    values: list of values of the index.
    Ex.
        cp.index("items",["Item 1","Item 2","Item 3"])
        cp.index("years",[2016,2017,2018])
    """
    if values is None:
        values = ["Item 1","Item 2","Item 3"]
    return cubepy.Index(name ,values)            


def find(param1, param2, compareType=1, caseSensitive = True):
    """
    param1: value or indexarray for compare
    param2: index compare to
    compareType: cp.exact=1, cp.start_with=2, cp.end_with=3, cp.contain=4  
    caseSensitive: able to differentiate between uppercase and lowercase (by default True)

    If param1 is a scalar (numeric or str) and param2 is an index:  return cube indexed by param2 with True on ocurrences of param2
        Ex. result = cp.apply("te", region, cp.end_with)
    If param1 is an index and param2 is an index too:  return cube indexed by param1 and param2 with True on ocurrences of param1 on param2
        Ex. result = cp.apply(subregion, region, cp.contain)

    """
    def _fn(item,value):
        if isinstance(item,str) == False:
            item = str(item)
        if isinstance(value,str) == False:
            value = str(value)
        if compareType==1:
            if caseSensitive:
                return item == value
            else:
                return item.lower() == value.lower()
        elif compareType==2:
            if caseSensitive:
                return item[:len(value)] == value
            else:
                return item[:len(value)].lower() == value.lower()                                    
        elif compareType==3:
            if caseSensitive:
                return item[-len(value):] == value
            else:
                return item[-len(value):].lower() == value.lower()                    
        elif compareType==4:
            if caseSensitive:
                return value in item
            else:
                return value.lower() in item.lower()

    if (isinstance(param1,str) or str(param1).isnumeric()) and isinstance(param2,cubepy.Index):
        vfn = np.vectorize(_fn)
        return cubepy.Cube([param2],vfn(param2.values,param1))

    if isinstance(param1,cubepy.Index) and isinstance(param2,cubepy.Index):
        _res = cubepy.Cube.full([param1,param2],False)
        rr=0
        for row in param1.values:
            cc=0
            for col in param2.values:
                _res.values[rr,cc] = _fn(col,row)
                cc+=1
            rr+=1
        return _res

def selectText(data, first = None, last = None ):
    """Returns a new cube with the text contained between the 'first' and 'last' characters of cube / index 'data'. Starts counting from 0.
            If 'first' character is ommited, it returns every character from the first character of 'data' to the 'last' character, inclusive.
            If 'last' character is ommited, it returns every character from "first" character of 'data', to the last character available for each cell.
    """
    if first is None:
        if last is None:
            sliced_data = apply( lambda x: x[:], data )
        else:
            sliced_data = apply( lambda x: x[:last], data )
    else:
        if last is None:
            sliced_data = apply( lambda x: x[first:], data )
        else:
            sliced_data = apply( lambda x: x[first:last], data )
    
    return sliced_data        
        

def apply(fn, param1, param2=None, start=None):
    """
    Apply functions to index and cubes. Multiple results can be obtained
    fn: function to apply
    param1: index or cube
    param2: index or cube
    start: scalar or cube
        Ex. 
            cp.apply(lambda x: x[:2] ,indexRegion): return new cube indexed by "indexRegion" and apply fn on each item
            cp.apply(lambda x: x*5 ,cubeSales): return new cube result of apply fn on all values of the cubeSales
            cp.apply( cp.addPeriods, start_proj , end_proj): Return new cube result of apply fn on "start_proj" with "end_proj"
            cp.apply(lambda x: x+1, indexYear, start=10) : Return new cube indexed by "indexYear", result of apply fn starting with scalar value "10"
            cp.apply(lambda x: x+1, indexYear, start=prices) : Return new cube indexed by "indexYear", result of apply fn starting with cube "prices"
    """
    if callable(fn):
        vfn = np.vectorize(fn)
        if param2 is None:
            if isinstance(param1,cubepy.Index):
                if start is None:
                    return cubepy.Cube([param1],vfn(param1.values)) 
                elif isinstance(start,cubepy.Cube):
                    values=[start.values]
                    numEls =len(param1)
                    for nn in range(numEls-1):
                        values.append(fn(values[nn]))
                    new_axes = start._axes.insert(param1, 0)
                    return cubepy.Cube(new_axes,values)
                else:
                    values=[start]
                    numEls =len(param1)
                    for nn in range(numEls-1):
                        values.append(fn(values[nn]))
                    return cubepy.Cube(param1,values)
            if isinstance(param1,cubepy.Cube):
                return param1.apply(fn)
        elif isinstance(param1,cubepy.Cube) and isinstance(param2,cubepy.Cube):
                return apply_op( param1 , param2 , vfn)
    return None


def max(cube1,cube2):
    """Return max value between two cubes
    """
    return (cube1 > cube2)*cube1 + (cube1 <= cube2) * cube2

def min(cube1,cube2):
    """Return min value between two cubes
    """
    return (cube1 > cube2)*cube2 + (cube1 <= cube2) * cube1

def sum(cube, axis=None, keep=None, group=None, sort_grp=True):
    """Sum of array elements over a given axis.
    :param axis: Axis or axes along which a sum is performed. The default (axis = None) is perform a sum
    over all the dimensions of the input array. axis may be negative, in which case it counts from the last
    to the first axis. If this is a tuple of ints, a sum is performed on multiple axes, instead of a single
    axis or all the axes as before.
    :return: new Cube instance or a scalar value
    """
    return cube.reduce(np.sum, axis, keep, group, sort_grp)    


def subscript(cube, index, value):
    """Filter cube1 using the index and the value. Return a new cube without the index1 dimension
        Ex.
            cp.subscript(nodo_ejemplo,index_para_ejemplo,"Item 1")
    """
    return cube[index==value]


def slice(cube, index, value):
    """Filter cube using the index and the value. Return a new cube without the index dimension
        Ex.
            cp.slice(nodo_ejemplo,index_para_ejemplo,2)
    """
    if isinstance(index,cubepy.Index):
        index = index.name
    if isinstance(value,str) or str(value).isnumeric():
        value =[value]
    return cube.take(index,value).squeeze()


def shift(cube, axis, n=1, cval=0):
    """Returns a cube with the axis shifted.
        Ex.
            cp.shift(nodo_ejemplo,index_para_ejemplo,1)
    """
    return cube.shift(axis,n,cval)


def subset(cube, indexName="new index"):
    """Returns a list of all the elements of the index for which cube is true. The function is used to create a new index that is a subset of an existing index.
        Ex. cp.subset(cantidades>0)
    """
    cond = cube>0
    values = cond.axes[0].values[cond.values]
    return index(indexName,values)


def aggregate(cube,mapCube,indexToRemove,targetIndex):
    """ Aggregates the values in Cube to generate the result indexed by  targetIndex.
        Map gives the value of targetIndex for each element of indexToRemove

        Example for aggregating time information into annual index the syntax is:
            cp.aggregate(cube, map, time, years )
    """
    grouping_index_mat = cubepy.Cube([targetIndex],targetIndex.values)
    mat_allocation = mapCube == grouping_index_mat
    return (cube * mat_allocation).sum(indexToRemove)


def cumulate(cube, index):
    """ TODO coment
    """
    pos=0
    tmpMat=cubepy.Cube.zeros(cube.axes)
    tmpInd=cubepy.Cube.zeros([index])
    for j in index.values:
        tmpInd.values[pos:pos+1]=1
        tmpMat = tmpMat + ((tmpInd * cube).sum(index))*(j==index)
        pos=pos+1
    return tmpMat


def cumProd(cube, index):
    """Return the cumulative product of elements along a given axis
        param cube: cube 
        param axis: axis name (str), index (int) or instance
        Ex:
            cp.cumProd(nodo,indice)
    """
    return cube.cumProd(index)


def irr(flow, time_index ):
    """Returns the Internal Rate of Return (IRR) of a series of periodic payments (negative values) and inflows (positive values). The IRR is the discount rate at which the Net Present Value (NPV) of the flows equals zero. 
        The variable flow must be indexed by time_index.

    If the cash flow never changes sign, cp.irr() has no solution and returns NAN (Not A Number).
    """
    import pandas as pd

    _cube_dimensions = index("flowdims",flow.dims ) 
    _rest_of_indexes_labels = subset( _cube_dimensions != time_index.name ).values 
    _rest_of_indexes = [flow.axis(xx) for xx in _rest_of_indexes_labels]

    _cube = None
    if len( _rest_of_indexes ) == 0: 
        _cube = np.irr( flow.values ) 
    else: 
        _cube = cube( _rest_of_indexes ) 
        _multivalues = [idx.values for idx in _rest_of_indexes] 

        _values = pd.MultiIndex.from_product( _multivalues ).values 

        for _item in _values: 
            _filter = []

            for _nn in range(len(_item)): 
                _filter.append( _rest_of_indexes[_nn].filter( [_item[_nn]] ) ) 

            _irr = np.irr( flow.filter( _filter ).squeeze().values ) 
            _cube.set_data( _filter, _irr )
    return _cube


def npv(rate, flow, time_index, offset = 1): 
    """"Returns the Net Present Value (NPV) of a cash flow with equally spaced periods. The flow parameter must contain a series of periodic payments (negative values) and inflows (positive values), indexed by time_index.
        The optional offset parameter especifies the offset of the first value relative to the current time period. By default, offset is set to 1, indicating that the first value is discounted as if it is one step in the future
    """
    _number_of_periods = time_index.pos + offset 
    _present_values = flow / ( 1 + rate ) ** _number_of_periods 
    _npv = _present_values.sum( axis = time_index ) 
    return _npv


def splitText(text,separator=",",resultIndex=None, indexName="new index"):
    """Split text into a list of substrings by each occurrence of separator.
        text:  string or cp.index used for split
        separator: string for use as separator
        resultIndex: optional, for return cp.cube indexes by this index
        indexName: optional, string name for new index
        Ex:
            cp.splitText("Uno Dos Tres Cuatro Cinco Seis", separator=" ") 
            cp.splitText(indice_para_split, separator="-") 
            cp.splitText("Uno Dos Tres Cuatro Cinco Seis"," ",result_index)  
            cp.splitText(indice_para_split, "-", result_index)  

    """
    if isinstance(text,str):
        _arr = text.split(separator)
        if resultIndex is None:
            return self.index(indexName,_arr)
        else:
            return self.cube(resultIndex,_arr)

    elif isinstance(text,cubepy.Index):
        _arr=[]
        for item in text.values:
            _tmpArr = str(item).split(separator)
            if not resultIndex is None:
                _tmpNpArray = np.array(_tmpArr)
                _tmpArr=[]
                for _nn in range(len(resultIndex)):
                    if (_nn<len(_tmpNpArray)):
                        _tmpArr.append(_tmpNpArray[_nn])
                    else:
                        _tmpArr.append("")

            _arr = _arr + _tmpArr
        _arrnp = np.array(_arr).flatten()
        
        if resultIndex is None:
            return index(indexName,_arrnp)
        else:
            return cube([text,resultIndex],_arrnp)


def changeIndex(cube, oldIndex, newIndex, compareMode=1,default=None):
    """Returns a new Cube instance with the axes changed,looking newIndex on oldIndex by value o by position.
        param Comparemode: cp.byVal or cp.byPos
        Ex:
            cp.changeIndex(nodo_original,indice_original,indice_cambiado,cp.byVal)
            cp.changeIndex(nodo_original,indice_original,indice_cambiado,cp.byPos)
    """
    return cube.change_axis(oldIndex,newIndex,compareMode,default=default)            


def copyIndex(cube, indexName="new index"):
    """Generate a cp.index with current unique values of the cube. 
    The cube must have only one dimension
    """
    import pandas as pd
    if cube.ndim>1:
        raise ValueError("The cube must have only one dimension")

    np_values = cube.values.flatten()
    seripandas = pd.Series(np_values)
    return index(indexName,seripandas.unique())


def cascadeVolume(demand,ranges,consumption_range_index=None):
    """TODO Comment
    Ex:
            cp.cascadeVolume(cantidades,limites_rango_consum)
            cp.cascadeVolume(cantidades,limites_rango_consum,"rangos_consumo")
            cp.cascadeVolume(cantidades,limites_rango_consum,rangos_consumo)
    """
    if consumption_range_index is None:
        if ranges.ndim==1:
            consumption_range_index = ranges.dims[0]
        else:
            raise ValueError("You must specify the consumption_range_index")
    elif isinstance(consumption_range_index,cubepy.Index):
        consumption_range_index = consumption_range_index.name


    vol_entre_lim = ranges - ranges.shift(consumption_range_index,-1,0)
    ConsumoBajoLimite = (ranges<demand)*vol_entre_lim
    # Forma de determinar el ultimo umbral de consumo
    ConsEnRangos = (demand>ranges)*1
    ultrango = (ConsEnRangos - ConsEnRangos.shift(consumption_range_index,-1,1))==-1
    ConsumoLimite = (demand - ConsumoBajoLimite.sum(consumption_range_index)) * ultrango
    return ConsumoBajoLimite + ConsumoLimite



def bandAllocation(demand,ranges,consumption_range_index=None):
    """TODO Comment
        Ex:
            cp.bandAllocation(cantidades,limites_rango_consum)
            cp.bandAllocation(cantidades,limites_rango_consum,"rangos_consumo")
            cp.bandAllocation(cantidades,limites_rango_consum,rangos_consumo)
    """

    if consumption_range_index is None:
        if ranges.ndim==1:
            consumption_range_index = ranges.dims[0]
        else:
            raise ValueError("You must specify the consumption_range_index")
    elif isinstance(consumption_range_index,cubepy.Index):
        consumption_range_index = consumption_range_index.name

    ConsEnRangos = (demand>ranges)*1
    ultrango = (ConsEnRangos - ConsEnRangos.shift(consumption_range_index,-1,1))==-1
    ConsumoLimite = demand * ultrango
    return ConsumoLimite


def dispatch(contract_vol, contract_price, contract_index, demand):
    """TODO comment
        Ex:
            cp.dispatch(contract_volumes,prices,contracts,demand)
    """
    order_index = index("order",range(len(contract_index)))
    contract_order = argsort(contract_price,contract_index)
    vol_by_dispatch_order = (contract_order==order_index)*contract_vol
    cum_vol = cumulate(vol_by_dispatch_order,order_index).sum(contract_index)
    return (cascadeVolume(demand,cum_vol,order_index)*(contract_order==order_index)).sum(order_index)


def argsort(cube,axis):
    """Return new cube with the position of values sorted by axis
        ex.
            cp.argsort(nodo_1,index_a)
    """
    return cube.argsort(axis)   


def iif(condition,truePart, falsePart=None):
    """Inline if. Evaluate condition and return truePart or fasePart
        ex. 
            cp.iif( producto=="Producto A", cantidades*1000 , cantidades)
            cp.iif( cantidades > 10, cantidades , 0)
    """
    if isinstance(condition,cubepy.Cube):
        _true  = condition * truePart
        if falsePart is None:
            return _true
        else:
            _false = (~condition) * falsePart
            return _true + _false
    elif isinstance(condition,bool):
        if condition:
            return truePart
        else:
            return falsePart
    return None    




def cubeFromPandas(dataframe, cubeIndexes, valueColumns, indexColumnHeaders=None,replaceByIndex=None):
    """Create new cp.cube, converting pandas to multidimensional data, according to the parameters
        dataframe: pandas dataframe
        cubeIndexes: objects cp.index for perform change index
        valueColumns: string with column name of the dataframe where contain the values
                        cp.index with columns names for convert colums to index
        indexColumnHeaders: (optional) column names in pandas to parse with cubeIndexes. Used if header on dataframe is not equal to index identifiers.
        replaceByIndex: (optional) replace index used in valueColumns for this index. (using changeindex)
    """
    import pandas as pd

    valueIndex=None
    if isinstance(valueColumns,cubepy.Index):
        valueIndex = valueColumns
        valueColumns = valueIndex.values
    elif isinstance(valueColumns,str):
        valueColumns = np.array([valueColumns])

    if indexColumnHeaders is None:
        indexColumnHeaders = [index.name for index in cubeIndexes]
        
    #create total index and index names

    _allindexes = cubeIndexes
    _allIndexNames = indexColumnHeaders[:]
    if not valueIndex is None:
        _allindexes.append(valueIndex)
        _allIndexNames.append("data_index")    

    #fill other columns for prevent melt error
    if isinstance(dataframe, pd.DataFrame):
        cols_not_in_df = [col for col in valueColumns if col not in dataframe.columns]
        for col in cols_not_in_df:
            dataframe[col] = np.nan


    _full = dataframe.reset_index().melt(id_vars=indexColumnHeaders,value_vars=valueColumns, var_name="data_index", value_name="data_value")

    if _full.size==0:
        _cube = cube(_allindexes, np.array([],dtype="O"))
    else:

        #sum for acum over duplicate records
        _full = _full.groupby(_allIndexNames, as_index=False).sum()
        _dtype = _full["data_value"].dtype
        
        _dataType = kindToString(_dtype.kind )
        if _dataType=="string":
            _full = _full[ (_full["data_value"]!="") & (_full['data_value'].notna()) ]
        else:
            _full = _full[ (_full["data_value"]!=0) & (_full['data_value'].notna()) ]


        _size = [len(x) for x in _allindexes]
        _emptyData = np.zeros(_size, dtype=_dtype )
        _cube = cube(_allindexes,_emptyData,_dtype)

        _valuePos = len(_full.columns)

        for _row in _full.itertuples():
            _arr=[]
            _isOK = True
            _value = _row[_valuePos]
            for nn in range(1,len(_allIndexNames)+1):
                _indexValue = _row[nn]
                if _indexValue in _allindexes[nn-1]._indices:
                    _pos = _allindexes[nn-1]._indices[_indexValue]
                    _arr.append(_pos)
                else:
                    _isOK=False
                    break

            if _isOK:    
                _cube._values[tuple(_arr)] = _value

        if (not valueIndex is None) and (not replaceByIndex is None):
            _cube = changeIndex(_cube, valueIndex,replaceByIndex, 2 )    

    return _cube 


def indexFromPandas(dataframe, columnName=None, removeEmpty=True, indexName="new index"):
    """ Return a cp.index from an column of a pandas dataframe.
    dataframe: pandas dataframe
    columnName: dataframe column name used for create cp.index. By default is created using the first column
    removeEmpty: True for remove empty rows
        Ex.
            cp.indexFromPandas(df)
            cp.indexFromPandas(df,"column10")
    """                
    
    _serie= None
    if columnName is None:
        _serie=dataframe[dataframe.columns[0]]
    else:
        _serie=dataframe[columnName]

    if removeEmpty:
        _serie.dropna(inplace=True)
        if kindToString( _serie.dtype.kind )=="string" or kindToString( _serie.dtype.kind )=="object":
            _serie = _serie[_serie!=""]

    return index(indexName,_serie.unique())


def pandasFromCube(cube, dataColumnName="value", keepIndexOrder=False ):
    """
    Return indexed Dataframe created from cubePy Cube
    """
    import pandas as pd

    multiIndexValues = [idx.values for idx in cube.axes]
    allIndexNames = [idx.name for idx in cube.axes]
    multiindex = pd.MultiIndex.from_product(multiIndexValues, names=allIndexNames)
    serie = pd.Series(data=cube.values.flatten() , index=multiindex)
    df = pd.DataFrame(serie)

    if keepIndexOrder:
        df = df.reset_index()
        for cubeIndex in cube.axes:
            df[cubeIndex.name] = df[cubeIndex.name].astype(pd.api.types.CategoricalDtype(cubeIndex.values))
        df.set_index(allIndexNames, inplace=True)
    df.columns= [dataColumnName]
    return df


def excel(filepath, useOpenpyxl=False, dataOnly=True, readOnly=True ):
    """ Create excel object from filepath.
    filepath: path to excel file
    useOpenpyxl: True for use custom 
    dataOnly: True for view only the values, not formula
    readOnly: True for read only, False for write options
    Ex.
            cp.excel("\path\to\the\excelfile.xlsx")
    """
    if isLinux():
        filepath = filepath.replace("\\","/")

    if os.path.isfile(filepath):
        if useOpenpyxl: # or isLinux():
            from openpyxl import load_workbook
            return load_workbook(filepath, data_only=dataOnly, read_only=readOnly)
        else:
            return filepath
    else:
        raise ValueError("File not found")    


def pandasFromExcel(excel,sheetName=None,namedRange=None,cellRange=None, indexes=None, driver='Driver={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)};DBQ=%s;READONLY=TRUE'):
    """ return a pandas dataframe from excel.
    excel: path to excel file or cp.excel object
    sheetName: sheet name to be read
    namedRange: name of the range to be read
    cellRange: used with sheetname, for read from a simple range
    indexes: Listo of columns names for convert to index of dataframe
        Ex.
            cp.pandasFromExcel(excelNode,"Sheet 1")
            cp.pandasFromExcel(excelNode,namedRange="name_range")
            cp.pandasFromExcel(excelNode,"Sheet 1",cellRange="A1:H10")

    """
    import pandas as pd 
    
    if isinstance(excel,str):

        if isLinux():
            filename=excel
            target_dir = os.path.dirname(filename)
            file_name, file_extension = os.path.splitext(filename)
            target_dir = os.path.join(target_dir, file_name)
            file_to_read_legacy = os.path.join(target_dir, f"{namedRange}.pkl")
            target_dir = f"{target_dir[:target_dir.rfind('/')+1]}.{target_dir[target_dir.rfind('/')+1:]}"

            file_to_read = os.path.join(target_dir, f"{namedRange}.pkl") if os.path.isfile(os.path.join(target_dir, f"{namedRange}.pkl")) else file_to_read_legacy

            if os.path.isfile(file_to_read):
                df = pd.read_pickle(file_to_read,compression='gzip')
                if not indexes is None:
                    df.set_index(indexes,inplace=True)
                return df
            else:
                #TODO: use openpyxl
                raise ValueError(f"Optimized version not found for: '{filename}'")

        else:
            import pyodbc

            cnxn = pyodbc.connect(driver % excel, autocommit=True)
            cursor = cnxn.cursor()
            table = ""
            if not sheetName is None: 
                table = "["+sheetName+"$]"
            if not namedRange is None:
                table = "["+namedRange+"]"
            if not cellRange is None:
                table = "["+sheetName+"$"+cellRange+"]"

            cursor.execute("SELECT * FROM " + table)
            rows = cursor.fetchall()
            columnNames = [str(x[0]) for x in cursor.description]
            cnxn.close()

            df = pd.DataFrame.from_records(rows,columns=columnNames)
            df = df.dropna(how ="all")
            if not indexes is None:
                df.set_index(indexes,inplace=True)
            return df
    else:
        
        if "openpyxl.workbook" in str(type(excel)):
            rangeToRead = None
            if not namedRange is None:
                the_range = excel.defined_names[namedRange]
                dests = the_range.destinations
                for title, coord in dests:
                    ws = excel[title]
                    rangeToRead=ws[coord]
            elif not cellRange is None:
                ws = excel[sheetName]
                rangeToRead = ws[cellRange]
            else: 
                rangeToRead = excel[sheetName]

            nn=0
            cols=[]
            values = []
            for row in rangeToRead:
                if nn==0:
                    cols = [str(c.value) for c in row]
                else:
                    values.append([c.value for c in row]) 
                nn+=1
            nn=0
            _finalCols=[]
            for _col in cols:
                if _col is None:
                    _finalCols.append("Unnamed" + str(nn))
                    nn+=1
                else:
                    _finalCols.append(_col)
            
            df = pd.DataFrame(values,columns=_finalCols)
            if not indexes is None:
                if isinstance(indexes,str):
                    indexes=[indexes]
                toIndex = []
                for indexColumn in indexes:
                    if indexColumn in df.columns.values:
                        toIndex.append(indexColumn)
                if len(toIndex)>0:
                    df.set_index(toIndex, inplace=True)

            return df.dropna(how ="all")
        else:
            raise ValueError("excel can be cp.excel object")        


def indexFromExcel(excel, sheetName=None,namedRange=None,cellRange=None, columnName=None, removeEmpty=True, indexName="new index"):
    """ Return a cp.index from an excel file.
    excel: cp.excel object
    sheetName: sheet name to be read
    namedRange: name of the range to be read
    cellRange: used with sheetname, for read from a simple range
    columnName: dataframe column name used for create cp.index. By default is created using the first column
    removeEmpty: True for remove empty rows
    indexName: new index name 
        Ex.
            cp.indexFromExcel(excelNode,"Sheet 1")
            cp.indexFromExcel(excelNode,namedRange="name_range")
            cp.indexFromExcel(excelNode,namedRange="name_range", columnName="indicadores")
    """                
    if isinstance(excel,str) or "openpyxl.workbook" in str(type(excel)):
        _df = pandasFromExcel(excel,sheetName,namedRange,cellRange)
        return indexFromPandas(_df,columnName,removeEmpty,indexName=indexName)
    else:
        raise ValueError("excel can be cp.excel object or a str path to the filename")


def cubeFromExcel(excel, sheetName=None,namedRange=None,cellRange=None, cubeIndexes=None, valueColumns=None, indexColumnHeaders=None, replaceByIndex=None):
    """ Return a cp.cube from excel file.
        excel: cp.excel object
    sheetName: sheet name to be read
    namedRange: name of the range to be read
    cellRange: used with sheetname, for read from a simple range
    cubeIndexes: objects cp.index for perform change index
    valueColumns: string with column name of the dataframe where contain the values
                    cp.index with columns names for convert colums to index
    indexColumnHeaders: (optional) column names in pandas to parse with cubeIndexes. Used if header on dataframe is not equal to index identifiers.
    replaceByIndex: (optional) replace index used in valueColumns for this index. (using changeindex)

        Ex.
            cp.cubeFromExcel(excelNode,"Sheet 1",cubeIndexes=[indicadores],valueColumns="descuentos")
            cp.cubeFromExcel(excelNode,namedRange="nombre_rango",cubeIndexes=[indicadores],valueColumns=time)
    """

    if isinstance(excel,str) or "openpyxl.workbook" in str(type(excel)):
        _df = pandasFromExcel(excel,sheetName,namedRange,cellRange)
        return cubeFromPandas(_df, cubeIndexes,  valueColumns, indexColumnHeaders, replaceByIndex=replaceByIndex) 
    else:
        raise ValueError("excel can be cp.excel object or a str path to the filename")


def sequence(initialNum, finalNum, stepSize = 1, dtype = None, indexName="new index"): 
    """
    Creates an index with a list of numbers increasing or decreasing from initial_num to final_num by increments (or decrements) of stepSize, which is optional and defaults to 1
    """
    values = np.arange(initialNum, finalNum + 1, step = stepSize, dtype = dtype) 
    return index(indexName,values)


def lookup(cubeWithData, cubeWithMap, sharedIndex): 
    """
    Returns the value of cubeWithData indexed by the index of cubeWithMap.
    cubeWithData must be indexed by sharedIndex and cubeWithData values must correspond to elements of sharedIndex.
    For example: Let's say you have a cube with an estimated inflation rate by Country ("inflation_rate" is the name of the cube; "country" is the name of the index) and you want to assign it to the corresponding Company depending on its location. On the other hand, there's a many-to-one map where each Company is allocated to a single Country ("country_to_company_allocation"). The sharedIndex, in this case, is Country ("country").
    As a result, 
        cp.lookup( inflation_rate , country_to_company_allocation , country )
    will return the estimated inflation rate by Company.
    """
    _final_cube = ((cubeWithMap == sharedIndex) * cubeWithData).sum(sharedIndex) 
    return _final_cube 



def cubeFromNumpy(npArray):
    """ 
    Return a cube object from numpy Array. Generate temporal indexes
    """
    _dimsNames = ["axis " + str(x) for x in range( npArray.ndim)]
    _dimsValues = [ list(x) for x in (range( npArray.shape[y]) for y in range( npArray.ndim)) ]
    _indexes = [cubepy.Index( _dimsNames[x], _dimsValues[x]) for x in range(len(_dimsNames))]
    _cube = cube(_indexes,npArray)
    return _cube


def movingStats(cube, index, window=12, fn=np.mean):
    """
        Allows to calculate in a moving Window the selected statistics measure
    """
    _realWindow = abs(window)
    _fInvert = lambda x: x[::-1].rolling(_realWindow).apply(fn)[::-1].shift(-1)
    _fNorm = lambda x: x.rolling(_realWindow).apply(fn).shift()
    _realF = _fNorm if window<0 else _fInvert
    _levelList = [lvl.name for lvl in cube.axes if lvl.name!=index.name]
    _df = pandasFromCube(cube)
    if cube.ndim>1:
        _df = _df.groupby(level=_levelList)["value"].apply(_realF)
    else:
        _df = _df.apply(_realF)
    return cubeFromPandas( _df, cube.axes, "value")


def isLinux():
    if platform == "linux" or platform == "linux2" or platform == "darwin":
        return True
    else:
        return False



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import jsonpickle


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "DataframeGroupby"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId):

            currentDef = model.getNode(nodeId).definition
            newDef = self.getLastDefinition(currentDef)
            newDef = newDef + "\n# Groupby"
            agg = params['agg']
            need_join = False

            group_by = ""
            if 'columns' in params and len(params['columns']) > 0:
                group_by = f".groupby({params['columns']})"
                # format aggregator
                for key in agg:
                    if len(agg[key]) == 1:
                        agg[key] = str(agg[key][0])
                    else:
                        need_join = True

            newDef = newDef + \
                f"\n_df = _df{group_by}.agg({jsonpickle.dumps(agg)})"
            if need_join:
                newDef = newDef + "\n# Format columns name"
                newDef = newDef + \
                    f"\n_df.columns = _df.columns.map('_'.join)"

            newDef = newDef + \
                f"\nresult = _df"

            model.getNode(nodeId).definition = self.formatDefinition(newDef)

            return "ok"

        return ""



from pyplan_core.cubepy.version import __version__
from pyplan_core.cubepy.axis import Axis
from pyplan_core.cubepy.cube import Cube, concatenate, stack, apply_op
from pyplan_core.cubepy.index import Index
from pyplan_core.cubepy.exceptions import *



class AxisAlignError(ValueError):
    """
    This error is raised when axis A is being aligned to axis B and:
    - A is Series:
      if any value on axis A does not equal to the corresponding value on axis B (order is significant)
      note that it does not matter if B is Index or Series
    - A is Index and B is Series:
      if B contains a value which is not contained in A
    - A and B are both Index objects and one of them contains a value which is not in the other one 
      (this is also the case when they have different lengths); 
      note that the order of the values can be arbitrary      
    """
    pass


class InvalidAxisLengthError(ValueError):
    pass


class NonUniqueDimNamesError(ValueError):
    pass



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
import pandas as pd
import jsonpickle


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "DataframeIndex"

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            currentDef = model.getNode(nodeId).definition
            newDef = self.getLastDefinition(currentDef)
            newDef = newDef + "\n# Set index"
            reset_index = ""
            if not isinstance(model.getNode(nodeId).result.index, pd.RangeIndex):
                reset_index = ".reset_index()"

            if not params is None and "columns" in params and len(params["columns"]) > 0:
                newDef = newDef + \
                    f"\nresult = _df{reset_index}.set_index({params['columns']})"
            else:
                newDef = newDef + f"\nresult = _df{reset_index}"

            model.getNode(nodeId).definition = self.formatDefinition(newDef)
            return "ok"
        return ""



default_formats = {
    "variable": {
        "color": "#4CBCFF",
        "nodeClass": "variable",
        "definition": "result = 0",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 1,
            "showLabel": 1,
            "showOutputs": 1,
            "useNodeFont": 0
        },
        "w": 96,
        "h": 48
    },
    "function": {
        "color": "#cb98ff",
        "nodeClass": "function",
        "definition": "def _fn(param1,param2):\n    return param1+param2\n\nresult = _fn",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 0,
            "showLabel": 1,
            "showOutputs": 0,
            "useNodeFont": 0
        },
        "h": 40,
        "w": 100
    },
    "decision": {
        "color": "#4cffa6",
        "nodeClass": "decision",
        "definition": "result = selector(['Item 1','Item 2','Item 3'],1,False)",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 1,
            "showLabel": 1,
            "showOutputs": 1,
            "useNodeFont": 0
        },
        "h": 50,
        "w": 100
    },
    "index": {
        "color": "#9999ff",
        "nodeClass": "index",
        "definition": "result = pd.Index(['Item 1','Item 2','Item 3'])",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 0,
            "showLabel": 1,
            "showOutputs": 0,
            "useNodeFont": 0
        },
        "w": 96,
        "h": 48
    },
    "button": {
        "color": "#CBCBCB",
        "nodeClass": "button",
        "definition": "",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 0,
            "showLabel": 1,
            "showOutputs": 0,
            "useNodeFont": 0
        },
        "h": 40,
        "w": 154
    },
    "module": {
        "color": "#9FC5E8",
        "nodeClass": "module",
        "definition": "",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 1,
            "showLabel": 1,
            "showOutputs": 1,
            "useNodeFont": 0
        },
        "h": 50,
        "w": 100
    },
    "text": {
        "color": "#EEEEEE",
        "nodeClass": "text",
        "definition": "",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 0,
            "showInputs": 0,
            "showLabel": 1,
            "showOutputs": 0,
            "useNodeFont": 0
        },
        "h": 164,
        "w": 376,
        "z": -1
    },
    "inputtable": {
        "color": "#ffff4b",
        "nodeClass": "inputtable",
        "definition": "result = 0",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 1,
            "showLabel": 1,
            "showOutputs": 1,
            "useNodeFont": 0
        },
        "w": 96,
        "h": 48
    },
    "sourcecsv": {
        "color": "#d9ead3",
        "nodeClass": "variable",
        "definition": "result = 0",
        "nodeInfo": {
            "fill": 1,
            "showBorder": 1,
            "showInputs": 1,
            "showLabel": 1,
            "showOutputs": 1,
            "useNodeFont": 0
        },
        "w": 96,
        "h": 48
    }
}



import json
import autopep8
from pandas import MultiIndex, DataFrame
from xarray import DataArray


class BaseWizard(object):

    def getLastDefinition(self, definition, variable="_df"):
        newDef = definition.replace(
            "result=", f"{variable} =").replace("result =", f"{variable} =")
        return newDef

    def getColumnList(self, model, params):
        res = []
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            nodeResult = model.getNode(nodeId).result
            if isinstance(nodeResult, DataFrame):
                sample_df = nodeResult.sample(
                    min(len(nodeResult), 10000)).reset_index()
                count_by_columns = None
                if "includeColumnList" in params and params["includeColumnList"]:
                    only_string = sample_df.select_dtypes(include="object")
                    serie_count = only_string.nunique()
                    serie_count = serie_count[serie_count < 500]
                    if len(serie_count) > 0:
                        count_by_columns = serie_count

                # append columns
                for nn, col in enumerate(list(sample_df.columns)):
                    if not col is None and not (col == "index" and not nodeResult.index is None and nodeResult.index.name is None):
                        item = dict(field=col, type="index" if col in nodeResult.index.names else "column",  dtype=self.kindToString(
                            sample_df.dtypes[nn].kind))
                        if not count_by_columns is None and col in count_by_columns:
                            item["values"] = sample_df[col].dropna(
                            ).sort_values().unique().tolist()
                        res.append(item)
        return res

    def getDataArrayDimensions(self, model, params):
        res = []
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            nodeResult = model.getNode(nodeId).result
            if isinstance(nodeResult, DataArray):
                for dim in nodeResult.dims:
                    dim_node = model.getNode(
                        dim) if model.existNode(dim) else None
                    title = dim_node.title if dim_node and dim_node.title else dim
                    item = dict(field=dim, title=title,
                                values=nodeResult.coords[dim].values.tolist())
                    res.append(item)
        return res

    def kindToString(self, kind):
        """Returns the data type on human-readable string
        """
        if kind in {'U', 'S'}:
            return "string"
        elif kind in {'b'}:
            return "boolean"
        elif kind in {'i', 'u', 'f', 'c'}:
            return "numeric"
        elif kind in {'m', 'M'}:
            return "date"
        elif kind in {'O'}:
            return "object"
        elif kind in {'V'}:
            return "void"

    def formatDefinition(self, definition):
        return autopep8.fix_code(definition)



import json

from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator
from bokeh.embed import json_item
from bokeh.plotting import Figure
from bokeh.layouts import LayoutDOM, layout
from bokeh.models import Widget


class BokehEvaluator(BaseEvaluator):

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        _result = result

        if isinstance(_result, Figure):  # create layout
            _result.sizing_mode = 'stretch_both'
        elif isinstance(result, Widget):
            pass
        elif isinstance(_result, LayoutDOM):  # update layout for all childrens
            _result.sizing_mode = 'stretch_both'
            for item in _result.children:
                if isinstance(item, LayoutDOM) and not isinstance(item, Widget):
                    if item.sizing_mode is None and item.width_policy == "auto" and item.height_policy == "auto":
                        item.sizing_mode = 'stretch_both'

        res = dict()
        res["result"] = "BOKEH:" + json.dumps(json_item(_result, "#bokeh-id#"))
        return json.dumps(res)





class BaseDynamic(object):


    @staticmethod
    def clearAllCircularDependency(stringDef):
        """ Replaces cp.dynamic(x,y,z) , pp.dynamic(x,y,z) and dynamic(x,y,z) for 0
        """
        response = stringDef
        initialIndex = -1
        startIndex = -1
        finalIndex = -1
        toReplace = ''
        dynLen = len('cp.dynamic(')
        initialIndex = stringDef.find('cp.dynamic(')
        if initialIndex == -1:
            initialIndex = stringDef.find('pp.dynamic(')
        if initialIndex == -1:
            initialIndex = stringDef.find('dynamic(')            
            dynLen = len('dynamic(')
        if initialIndex != -1:
            startIndex = initialIndex
            initialIndex = initialIndex + dynLen
            if len(stringDef) > initialIndex:
                finalIndex = stringDef[initialIndex+1:].find(')')

        if initialIndex != -1 and finalIndex != -1:
            toReplace = stringDef[startIndex:initialIndex + finalIndex + 2]
            response = stringDef.replace(toReplace, "0")

        return response


# implementations

    def circularEval(self, node, params): raise ValueError("not implemented")

    def generateLoopDef(self, node, nodeDefinition,
                        cyclicVariables): raise ValueError("not implemented")

    def generateInitDef(self, node, nodeResult,
                        dynamicIndex): raise ValueError("not implemented")

    def generateCircularParameters(
        self, node, nodeList): raise ValueError("not implemented")

    def clearCircularDependency(
        self, stringDef, replaceWith="0"): raise ValueError("not implemented")



import json

import numpy as np
import pandas as pd
import xarray as xr
from pyplan_core.classes.evaluators.XArrayEvaluator import XArrayEvaluator
from pyplan_core.classes.common.filterChoices import filterChoices
from pyplan_core.classes.common.indexValuesReq import IndexValuesReq
from pyplan_core.cubepy.cube import kindToString


class NumpyEvaluator(XArrayEvaluator):

    AXISNAME = "Axis "

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        cube = self.createCube(nodeDic[nodeId].identifier, result)
        return super().evaluateNode(cube, nodeDic, nodeId, dims, rows, columns, summaryBy, bottomTotal, rightTotal, fromRow, toRow)

    def hasDim(self, result, dim):
        return True if dim.split(".")[0] in result.dims else False

    def getIndexes(self, node, result=None):
        res = []
        if result is None:
            result = node._result

        res = [self.AXISNAME +
               str(x)+"."+node.identifier for x in range(result.ndim)]
        return res

    def getIndexesWithLevels(self, node, result=None):
        res = []
        if result is None:
            result = node._result

        if not result is None:
            for x in range(result.ndim):
                item = {"field": self.AXISNAME + str(x)+"."+node.identifier, "name": self.AXISNAME + str(
                    x), "description": "", "levels": []}
                res.append(item)
        return res

    def getIndexValues(self, nodeDic, data: IndexValuesReq, result=None):
        res = []
        if data.node_id:
            if (not data.node_id is None) & (data.node_id in nodeDic):
                node = nodeDic[data.node_id]
                if result is None:
                    result = node.result
                axe = int(data.index_id.split(" ")[1])
                lenAxe = result.shape[axe]
                res = list(range(lenAxe))
        if data.text1:
            text1 = data.text1.lower()
            if data.filter == filterChoices.CONTAINS.value:
                res = list(
                    filter(lambda item: text1 in str(item).lower(), res))
            elif data.filter == filterChoices.NOT_CONTAINS.value:
                res = list(
                    filter(lambda item: not text1 in str(item).lower(), res))
        return res

    def addToFilter(self, nodeDic, dim, filters):
        if "values" in dim and dim["values"] is not None and len(dim["values"]) > 0:
            field = str(dim["field"]).split(".")[0]
            # chek if the node id of the index we are using to filter has changed

            nodeId = None
            indexType = self.getIndexType(nodeDic, nodeId, field)

            # check if the indexes have change
            _values = None
            if indexType == "S":
                _values = [str(xx["value"]) for xx in dim["values"]]
            else:
                _values = [int(xx["value"]) for xx in dim["values"]]

            all_values = None
            npValues = np.array(_values)
            if len(npValues) > 0:
                filters[field] = npValues

    def getIndexType(self, nodeDic, nodeId, indexId):
        return "N"

    def getCubeValues(self, result, nodeDic, nodeId, query):
        cube = self.createCube(nodeDic[nodeId].identifier, result)
        return super().getCubeValues(cube, nodeDic, nodeId, query)

    def getCubeDimensionValues(self, result, nodeDic, nodeId, query):
        cube = self.createCube(nodeDic[nodeId].identifier, result)
        return super().getCubeDimensionValues(cube, nodeDic, nodeId, query)

    def getCubeMetadata(self, result, nodeDic, nodeId):
        cube = self.createCube(nodeDic[nodeId].identifier, result)
        return super().getCubeMetadata(cube, nodeDic, nodeId)

    def createCube(self, nodeId, npArray):
        _dimsNames = [self.AXISNAME + str(x) for x in range(npArray.ndim)]
        _dimsValues = [list(x) for x in (range(npArray.shape[y])
                                         for y in range(npArray.ndim))]
        _indexes = [pd.Index(_dimsValues[x], name=_dimsNames[x])
                    for x in range(len(_dimsNames))]

        data_array = xr.DataArray(npArray, _indexes)
        return data_array

    def isTable(self, node):
        res = "0"
        if isinstance(node.result, np.ndarray):
            if not node.definition is None and node.definition != "":
                import re
                deff = re.sub(
                    '[\s+]', '', str(node.definition).strip(' \t\n\r')).lower()
                if (deff.startswith("result=np.")):
                    res = "1"
        return res

    def setNodeValueChanges(self, nodeDic, nodeId, nodeChanges):
        result = nodeDic[nodeId].result
        for change in nodeChanges["changes"]:
            newValue = change["definition"]
            slices = [slice(None)] * nodeDic[nodeId].result.ndim
            for filterItem in change["filterList"]:
                pos = int(filterItem["Value"])
                axis = int(
                    str(str(filterItem["Key"]).split(".")[0]).split(" ")[1])
                slices[axis] = pos

            result[tuple(slices)] = newValue

        np.set_printoptions(threshold=np.prod(result.shape))
        data = np.array2string(result, separator=",").replace('\n', '')
        newDeff = "result = np.array(" + data + ")"
        nodeDic[nodeId].definition = newDeff
        return "ok"

    def previewNode(self, nodeDic, nodeId):
        from pyplan_core.cubepy.Helpers import Helpers
        from sys import getsizeof

        result = nodeDic[nodeId].result
        res = {
            "resultType": str(type(result)),
            "dims": [],
            "console": nodeDic[nodeId].lastEvaluationConsole,
            "preview": ""
        }

        for nn in range(result.ndim):
            _item = self.AXISNAME+str(nn) + " [" + str(result.shape[nn]) + "]"
            res["dims"].append(_item)

        res["preview"] += "Dimensions: " + str(result.ndim)
        res["preview"] += "\nShape: " + str(result.shape)
        res["preview"] += "\nSize: " + str(result.size)
        res["preview"] += "\nMemory: " + \
            str(round(result.nbytes/1024/1024, 2)) + " Mb"
        res["preview"] += "\nData type: " + \
            str(result.dtype) + " (" + kindToString(result.dtype.kind)+")"
        res["preview"] += "\nValues: \n\n" + str(result)[:1000]

        result = None

        return json.dumps(res)



import numpy as np

from pyplan_core.cubepy.axis import Axis


class Index(Axis):
    """A named sequence of unique indexed values. Can be used as indexable axis in Cube.
    Name is a string. Values are stored in one-dimensional numpy array.
    """

    def __init__(self, name, values):
        """Initialize a new Index object. The values must be unique, otherwise ValueError is raised.
        :param name: str
        :param values: sequence of values (must be unique), 
                        sequence of cubepy.Index 
        :raise: ValueError if there are duplicate values
        """

        if isinstance(values, list) and len(values) > 0 and isinstance(values[0], self.__class__):
            nn = 0
            for idx in values:
                if nn == 0:
                    fullValues = np.copy(idx.values)
                else:
                    fullValues = np.concatenate(
                        (fullValues, idx.values), axis=0)
                nn += 1
            values = np.unique(fullValues)

        super(Index, self).__init__(name, values)

        # create dictionary
        self._indices = {x: i for i, x in enumerate(self._values)}

        # values must not be change once the index has been created
        self._values.flags.writeable = False

        if len(self._indices) != len(self._values):
            raise ValueError('Index cannot have duplicate values')

        self._vectorized_index = np.vectorize(
            self._indices.__getitem__, otypes=[np.int])
        self._vectorized_contains = np.vectorize(
            self._indices.__contains__, otypes=[np.bool])

    def __contains__(self, item):
        """Implementation of 'in' operator.
        :param item: a value to be looked up whether exists
        :return: bool
        """
        return item in self._indices

    def contains(self, item):
        """Tests whether item or items exist among values.
        If item is single value, then return a single boolean value.
        If item is a sequence, then return numpy array of booleans.
        :param item: a single value or a sequence of values
        :return: bool or numpy array of bools
        """
        v = self._vectorized_contains(item)
        if v.ndim > 0:
            return v
        return v.item()

    def indexof(self, item):
        """If item is single value, then return a single integer value.
        If item is a sequence, then return numpy array of integers.
        :param item: a single value or a sequence of values
        :return: int or numpy array of ints
        :raise: KeyError if value does not exist
        """
        v = self._vectorized_index(item)
        if v.ndim > 0:
            return v
        return v.item()

    @property
    def pos(self):
        from pyplan_core.cubepy.cube import Cube
        return Cube([self], range(len(self)))



from pyplan_core.classes.wizards.BaseWizard import BaseWizard
from pyplan_core.classes.evaluators.XArrayEvaluator import XArrayEvaluator
import xarray as xr
import json


class Wizard(BaseWizard):

    def __init__(self):
        self.code = "InputTable"

    def getNodeSchema(self, model, params):
        nodeId = params["nodeId"]

        res = dict()
        res["model_indexes"] = [{"title": node.title if node.title else "",
                                 "id": node.identifier} for node in model.findNodes('nodeClass', "index")]
        res["model_indexes"].sort(key=lambda x: x["id"])

        res["default_value"] = "0."
        res["node_indexes"] = []
        if model.existNode(nodeId):
            base_node = model.getNode(nodeId)
            da = base_node.result
            if isinstance(da, xr.DataArray):
                # get default value
                if "_input_properties" in base_node.definition:
                    _input_properties = model.evaluate(base_node.definition.split("# values")[
                                                       0] + "result = _input_properties")
                    res["default_value"] = str(_input_properties["defaultValue"]) if str(
                        _input_properties["defaultValue"]) else "''"

                model_indexes = res["model_indexes"]
                node_indexes = res["node_indexes"]
                for dim in da.dims:
                    node_index = next(
                        (index for index in model_indexes if index["id"] == dim), None)
                    model_indexes.remove(node_index)
                    node_indexes.append(node_index)
        return res

    def generateDefinition(self, model, params):
        nodeId = params["nodeId"]
        if model.existNode(nodeId):
            base_node = model.getNode(nodeId)

            default_value = params["default_value"]
            if default_value:
                try:
                    default_value = int(default_value)
                except:
                    try:
                        default_value = float(default_value)
                    except:
                        if default_value in ["True", "False"]:
                            default_value = default_value == "True"
                        else:
                            default_value = default_value.strip(
                                "\"").strip("\'")

                if str(default_value).isnumeric():
                    default_value = int(default_value)

                new_indexes = params["node_indexes"]

                da = base_node.result
                if not isinstance(da, xr.DataArray):
                    da = xr.DataArray(default_value, coords=[])

                # remove unused dims
                for dim in da.dims:
                    if not dim in new_indexes:
                        da = da.isel({dim: 0}).drop(dim).squeeze()

                # add new dims
                for new_index in new_indexes:
                    if not new_index in da.dims:
                        da = da.expand_dims(
                            {new_index: list(model.getNode(new_index).result.values)})

                # generate new definition
                evaluator = XArrayEvaluator()
                base_node._result = da
                new_def = evaluator.generateNodeDefinition(
                    model.nodeDic, nodeId, defaultValue=default_value)

                # test def:
                temp_res = model.evaluate(new_def)
                if not isinstance(temp_res, xr.DataArray):
                    raise ValueError(
                        f"New definition for node is not DataArray. New type: {type(temp_res)} ")
                base_node.definition = new_def
            else:
                raise ValueError("Default value can't be empty")
            return "ok"

        return ""



import json


class NodeInfo(dict):

    def __init__(self, showInputs, showOutputs, showLabel, showBorder, fill, useNodeFont):
        """
        self.showInputs = showInputs
        self.showOutputs = showOutputs
        self.showLabel = showLabel
        self.showBorder = showBorder
        self.fill = fill
        self.useNodeFont = useNodeFont
        """

        dict.__init__(self, showInputs=showInputs, showOutputs=showOutputs,
                      showLabel=showLabel, showBorder=showBorder, fill=fill, useNodeFont=useNodeFont)

    def __setattr__(self, name, value):
        if name == "fullset":
            arr = value.split(",")
            if (len(arr) > 6):
                self["showInputs"] = 1 if arr[1] == "1" else 0
                self["showOutputs"] = 1 if arr[2] == "1" else 0
                self["showLabel"] = 1 if arr[3] == "1" else 0
                self["showBorder"] = 1 if arr[4] == "1" else 0
                self["fill"] = 1 if arr[5] == "1" else 0
                self["useNodeFont"] = 1 if arr[6] == "1" else 0
        else:
            self[name] = value

    def __del__(self):
        self.release()

    def __repr__(self):
        return json.dumps(self.__dict__)

    def release(self):
        pass

    @property
    def showInputs(self):
        return self["showInputs"] == 1

    @property
    def showOutputs(self):
        return self["showOutputs"] == 1

    @property
    def showLabel(self):
        return self["showLabel"] == 1

    @property
    def showBorder(self):
        return self["showBorder"] == 1

    @property
    def fill(self):
        return self["fill"] == 1

    @property
    def useNodeFont(self):
        return self["useNodeFont"] == 1



import numpy as np

class Axis(object):
    """A named sequence of values. Can be used as non-indexable axis in Cube.
    Name is a string. Values are stored in one-dimensional numpy array.
    """
    

    def __init__(self, name, values):
        """Initializes Axis object.
        :param name: str
        :param values: sequence of values of the same type, are converted to 1-D numpy array
        :raise: ValueError if values cannot be converted, TypeError if name is not string
        """
        if not isinstance(name, str):
            raise TypeError("type of {} is not str".format(repr(name)))
        values = np.atleast_1d(values)
        if values.ndim > 1:
            raise ValueError("values must not have more than 1 dimension")
        self._name = name
        self._values = values

    def __repr__(self):
        """Returns textual representation of Axis object. Can be reused by inherited classes.
        :return: str
        """
        return "{}('{}', {})".format(self.__class__.__name__, self._name, self._values)
        
    def __len__(self):
        """Returns the number of elements in (the length) the axis.
        :return: int
        """
        return len(self._values)

    def __getitem__(self, item):
        """
        :param item:
        :return: a new Axis object
        """
        _newValues = self._values[item]
        if isinstance(_newValues,str):
            _newValues = [_newValues]
        elif "numpy." in str(type(_newValues)):
            _newValues=_newValues.flatten()

        return self.__class__(self._name, _newValues)


    def __sizeof__(self):
        return self.values.nbytes

    # A == B
    def __eq__(self, other):
        return apply_op(self,other, np.ndarray.__eq__)

    # A != B
    def __ne__(self, other):
        return apply_op(self,other, np.ndarray.__ne__)    

    # A < B
    def __lt__(self, other):
        return apply_op(self,other, np.ndarray.__lt__)        

    # A <= B
    def __le__(self, other):
        return apply_op(self,other, np.ndarray.__le__)            

    # A > B
    def __gt__(self, other):
        return apply_op(self,other, np.ndarray.__gt__)            

    # A >= B
    def __ge__(self, other):
        return apply_op(self,other, np.ndarray.__ge__)            


    @property
    def name(self):
        """Returns he name of the axis.
        :return: str
        """
        return self._name

    @property
    def values(self):
        """Returns one-dimensional numpy.ndarray of axis values.
        :return: numpy.ndarray
        """
        return self._values  # TODO: view?

    def filter(self, values):
        """Filter axis elements which are contained in values. The axis order is preserved.
        :param values: a value or a list, set, tuple or numpy array of values
            the order or values is irrelevant, need not be unique
        :return:
        """
        if isinstance(values, set):
            values = list(values)
        values = np.asarray(values)
        selection = np.in1d(self._values, values)
        return self.__class__(self._name, self._values[selection])

    def like(self,text):
        """Return Cube with  this index and True when elemen contains text
        """
        from pyplan_core.cubepy.cube import Cube
        
        _vals = []
        for _value in self.values:
            if text in str(_value):
                _vals.append(True)
            else:
                _vals.append(False)
        return Cube( [self],_vals)

        
    def take(self, indices):
        """Analogy to numpy.ndarray.take.
        :return: a new Axis object
        """
        return self.__class__(self._name, self._values.take(indices))
        
    def compress(self, condition):
        """Analogy to numpy.ndarray.compress.
        :return: a new Axis object
        """
        return self.__class__(self._name, self._values.compress(condition))

    def rename(self, new_name):
        """Returns a new object (of type Axis or the actual derived type) with the new name and the same values.
        :param new_name: str
        :return: new axis (instance of actual derived type)
        """
        return self.__class__(new_name, self._values)

    def sort(self):
        """Sorts the values.
        :return: a new Axis object with sorted values
        """
        return self.__class__(self._name, np.sort(self._values))

    def astype(self,dtype):
        """ change dtype of values of index """
        return self.__class__(self._name,self.values.real.astype(dtype,copy=False))

    def apply(self, func, *args):
        """Applies a function to each element individually and return the new index.
        :param func: function to be applied to values
        :param args: additional optional arguments of func
        :return: new Index  instance

        Examples:
        cube.apply(lambda x: x[0:5] )
        """

        # now with vectorize for ever
        func = np.vectorize(func)
        np_values = func(self.values, *args)
        seripandas = pd.Series(np_values)
        return self.__class__(self._name,seripandas.unique())



    def generateDefinition(self):
        """Generate code for inex definition"""
        #indexes = str([idx.name for idx in self.axes]).replace("'",'')
        np.set_printoptions(threshold = np.prod(self.values.shape))
        data = np.array2string(self.values, separator=",").replace('\n','')
        deff = "result = cp.index(" + data + ")"
        return deff




def apply_op(a,b, fn, *args):
    from pyplan_core.cubepy.cube import Cube

    if isinstance(a,Axis) and isinstance(b,Axis):
        return  getattr(Cube([a],a.values),fn.__name__)(Cube([b],b.values))
    elif isinstance(a,Axis):
        return Cube( [a],fn(a.values,b, *args))
    else:
        #TODO: Comparar tipos para mostrar error mas especifico
        return Cube([b], fn(b.values,a, *args))
            


import json

from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator


class IPythonEvaluator(BaseEvaluator):
    """ IPython Evaluator. 
    Display an IPython component using _repr_*_ methods
    """

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        return self.toHTML(result)

    def previewNode(self, nodeDic, nodeId):
        res = {
            "resultType": str(type(nodeDic[nodeId].result)),
            "dims": [],
            "columns": [],
            "console": nodeDic[nodeId].lastEvaluationConsole,
            "preview": self.toHTML(nodeDic[nodeId].result)
        }
        return json.dumps(res)

    def toHTML(self, result):
        _dir = dir(result)
        if "_repr_html_" in _dir:
            return result._repr_html_()
        elif "_repr_pretty_" in _dir:
            return result._repr_pretty_()
        elif "_repr_png_" in _dir:
            return result._repr_png_()
        elif "_repr_jpeg_" in _dir:
            return result._repr_jpeg_()
        elif "_repr_json_" in _dir:
            return result._repr_json_()



from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator
import json
from io import BytesIO
import base64
from matplotlib.figure import Figure as MatplotlibFigure
import inspect
import matplotlib.pyplot as plt


class MatplotlibEvaluator(BaseEvaluator):

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        fig = result
        _img = ""
        if not isinstance(fig, MatplotlibFigure):
            fig = plt

        figfile = BytesIO()
        fig.savefig(figfile, format='png')
        figfile.seek(0)  # rewind to beginning of file
        figdata_png = base64.b64encode(figfile.getvalue()).decode('utf8')
        _img = f"<img class='pyplan-autosize' src='data:image/png;base64,{figdata_png}'\>"
        nodeDic[nodeId]._result = _img
        plt.figure()

        res = dict()
        res["result"] = _img
        return json.dumps(res)



import json
import math

import numpy as np
import pandas as pd
import xarray as xr

from pyplan_core.classes.evaluators.BaseEvaluator import BaseEvaluator
from pyplan_core.classes.evaluators.PandasEvaluator import PandasEvaluator
from pyplan_core.classes.common.filterChoices import filterChoices
from pyplan_core.classes.common.indexValuesReq import IndexValuesReq


class XArrayEvaluator(BaseEvaluator):

    PAGESIZE = 100
    MAX_COLUMS = 5000

    def evaluateNode(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        if isinstance(result, xr.DataArray):
            return self.cubeEvaluate(result, nodeDic, nodeId, dims, rows, columns, summaryBy, bottomTotal, rightTotal, fromRow, toRow)

    def cubeEvaluate(self, result, nodeDic, nodeId, dims=None, rows=None, columns=None, summaryBy="sum", bottomTotal=False, rightTotal=False, fromRow=0, toRow=0):
        result_structure = self.getStructure(result)
        sby = np.nansum
        if summaryBy == 'avg':
            sby = np.nanmean
        elif summaryBy == 'max':
            sby = np.nanmax
        elif summaryBy == 'min':
            sby = np.nanmin

        if (fromRow is None) or int(fromRow) <= 0:
            fromRow = 1
        if (toRow is None) or int(toRow) < 1:
            toRow = 100
        fromRow = int(fromRow)
        toRow = int(toRow)

        result = self.applyHierarchy(
            result, nodeDic, nodeId, dims, rows, columns, sby)

        _filters = {}
        _rows = []
        _columns = []
        if not rows is None:
            for row in rows:
                if self.hasDim(result, str(row["field"])):
                    _rows.append(str(row["field"]).split(".")[0])
                    self.addToFilter(nodeDic, row, _filters)

        if not columns is None:
            for column in columns:
                if self.hasDim(result, str(column["field"])):
                    _columns.append(str(column["field"]).split(".")[0])
                    self.addToFilter(nodeDic, column, _filters)

        if not dims is None:
            for dim in dims:
                if self.hasDim(result, str(dim["field"]).split(".")[0]):
                    self.addToFilter(nodeDic, dim, _filters)

        tmp = None
        filteredResult = result
        if len(_filters) > 0:
            filteredResult = result.sel(_filters)

        if len(_rows) == 0 and len(_columns) == 0 and result.ndim > 0:
            try:
                tmp = sby(filteredResult)
            except Exception as ex:
                if "flexible type" in str(ex):
                    tmp = sby(filteredResult.astype("O"))
                else:
                    raise ex
            # fix np.nansum don't return dataarray 0 dimensional like np.sum
            if not isinstance(tmp, xr.DataArray):
                tmp = xr.DataArray(tmp)

        else:
            otherDims = [
                xx for xx in filteredResult.dims if xx not in (_rows + _columns)]
            if len(otherDims) > 0:
                squeezable = [filteredResult.sizes[xx]
                              == 1 for xx in otherDims]
                to_squeeze = list(np.array(otherDims)[squeezable])
                if len(to_squeeze) > 0:
                    filteredResult = filteredResult.squeeze(to_squeeze)
                    otherDims = [
                        xx for xx in filteredResult.dims if xx not in (_rows + _columns)]

            if len(otherDims) > 0:
                try:
                    tmp = filteredResult.reduce(
                        sby, otherDims).transpose(*(_rows + _columns))
                except Exception as ex:
                    if "flexible type" in str(ex):
                        tmp = filteredResult.astype("O").reduce(
                            sby, otherDims).transpose(*(_rows + _columns))
            else:
                tmp = filteredResult.transpose(*(_rows + _columns))

        finalValues = tmp.values
        finalIndexes = []
        if tmp.ndim > 0:
            finalIndexes = tmp.coords[tmp.dims[0]].values
        finalColumns = ["Total"]
        if tmp.ndim == 2:
            finalColumns = tmp.coords[tmp.dims[1]].values

        # Add totales
        _totalRow = None
        if bottomTotal and len(_rows) > 0:
            # add total row
            if tmp.ndim == 1:
                _totalRow = finalValues.sum(axis=0).reshape(1)
            else:
                _totalRow = finalValues.sum(
                    axis=0).reshape(1, len(finalValues[0]))
                _totalRow = _totalRow[0]
                if rightTotal:
                    _totalRow = np.append(_totalRow, finalValues.sum())

        if rightTotal and len(_columns) > 0:
            # add total column
            if tmp.ndim == 1:
                finalIndexes = np.append(finalIndexes, "Total")
                finalValues = np.append(
                    finalValues, finalValues.sum(axis=0).reshape(1), axis=0)
            else:
                finalColumns = np.append(finalColumns, "Total")
                finalValues = np.append(finalValues, finalValues.sum(
                    axis=1).reshape(len(finalValues), 1), axis=1)

        # chek inf
        if self.kindToString(finalValues.dtype.kind) == "numeric":
            if np.isinf(finalValues).any():
                finalValues[np.isinf(finalValues)] = None

        # chec if haver nan values
        if pd.isnull(finalValues).any():
            try:
                finalValues = np.where(
                    np.isnan(finalValues), None, finalValues)
            except:
                finalValues[pd.isnull(finalValues)] = None

        res = {}
        pageInfo = None
        onRow = None
        onColumn = None
        if len(_rows) == 0 and len(_columns) == 0:
            res = {
                "columns": [],
                "index": ["Total"],
                "data": [[finalValues.tolist()]]
            }
        elif len(_rows) == 0:
            onColumn = _columns[0]
            res = {
                "columns": self.checkDateFormat(finalIndexes[:XArrayEvaluator.MAX_COLUMS]).tolist(),
                "index": finalColumns,
                "data": [finalValues[:XArrayEvaluator.MAX_COLUMS].tolist()]
            }
        elif len(_columns) == 0:

            if (len(finalIndexes) > self.PAGESIZE):
                pageInfo = {
                    "fromRow": int(fromRow),
                    "toRow": int(toRow),
                    "totalRows": len(finalIndexes)
                }

            onRow = _rows[0]
            res = {
                "columns": finalColumns,
                "index": self.checkDateFormat(finalIndexes[fromRow-1:toRow]).tolist(),
                "data": [[x] for x in finalValues[fromRow-1:toRow].tolist()]
            }
            # add total rows
            if not _totalRow is None:
                res["index"].append("Total")
                res["data"].append(_totalRow.tolist())

        else:
            onColumn = _columns[0]
            onRow = _rows[0]

            if (len(finalIndexes) > self.PAGESIZE):
                pageInfo = {
                    "fromRow": int(fromRow),
                    "toRow": int(toRow),
                    "totalRows": len(finalIndexes)
                }

            res = {
                "columns": self.checkDateFormat(finalColumns[:XArrayEvaluator.MAX_COLUMS]).tolist(),
                "index": self.checkDateFormat(finalIndexes[fromRow-1:toRow]).tolist(),
                "data": finalValues[fromRow-1:toRow, :XArrayEvaluator.MAX_COLUMS].tolist()
            }

            # add total rows
            if not _totalRow is None:
                res["index"].append("Total")
                res["data"].append(
                    _totalRow[:XArrayEvaluator.MAX_COLUMS].tolist())

        return self.createResult(res, result_structure, onRow=onRow, onColumn=onColumn, node=nodeDic[nodeId], pageInfo=pageInfo)

    def getStructure(self, result):
        structure = dict()
        structure["type"] = str(type(result))
        structure["dims"] = list(result.dims)
        return structure

    def checkStructure(self, result, resultType):
        """ Check current vs result structure. Result False for distinct structure """
        res = True
        if resultType:
            try:
                structure = json.loads(resultType)
                result_structure = self.getStructure(result)
                res = structure["type"] == result_structure["type"] and all(elem in list(result_structure["dims"])
                                                                            for elem in list(structure["dims"]))
            except Exception as ex:
                print(f"Error checking structure: {ex}")

        return res

    def hasDim(self, result, dim):
        return True if dim.split(".")[0] in result.dims else False

    def addToFilter(self, nodeDic, dim, filters):
        if "values" in dim and dim["values"] is not None and len(dim["values"]) > 0:
            field = str(dim["field"]).split(".")[0]
            # chek if the node id of the index we are using to filter has changed

            nodeId = None
            indexType = None
            indexType = self.getIndexType(nodeDic, nodeId, field)

            # check if the indexes have change
            _values = None
            if indexType == "S":
                _values = [str(xx["value"]) for xx in dim["values"]]
            else:
                _values = [int(xx["value"]) for xx in dim["values"]]

            all_values = None
            npValues = np.array(_values)
            if field in nodeDic:
                all_values = nodeDic[field].result.values
            elif len(dim["field"].split(".")) > 1:
                node_id = str(dim["field"]).split(".")[1]
                if field in nodeDic[node_id].result.dims:
                    all_values = nodeDic[node_id].result.coords[field].values
            serie = pd.Series(all_values)
            if not all_values is None and serie.isin(npValues).any():
                npValues = all_values[serie.isin(npValues)]

            if len(npValues) > 0:
                filters[field] = npValues

    def getIndexes(self, node, result=None):
        if result is None:
            result = node._result
        return [(xx+"."+node.identifier) for xx in result.dims]

    def getIndexesWithLevels(self, node, result=None):
        res = []
        if result is None:
            result = node._result

        if not result is None:
            _model = node.model
            for indexItem in result.dims:
                itemDim = indexItem.split(",")[0]
                item = {"field": itemDim+"."+node.identifier,
                        "name": itemDim, "description": "", "levels": []}

                if _model.existNode(itemDim):
                    levelNode = _model.getNode(itemDim)
                    if levelNode.title:
                        item["name"] = levelNode.title
                        item["description"] = levelNode.description
                    if levelNode.numberFormat:
                        item["numberFormat"] = levelNode.numberFormat

                    # check for levels

                    if not levelNode.hierarchy_parents is None:

                        def buildLevels(parents, levelList):
                            if not isinstance(parents, list):
                                parents = [parents]
                            for parentIndexId in parents:
                                parentIndexNode = _model.getNode(parentIndexId)
                                if parentIndexNode is None:
                                    raise ValueError(
                                        f"Node {parentIndexId} not found")

                                levelItem = {
                                    "field": parentIndexId, "name": parentIndexNode.title or parentIndexId}
                                levelList.append(levelItem)
                                _dummy = parentIndexNode.result  # to force calc
                                if not parentIndexNode.hierarchy_parents is None:
                                    buildLevels(
                                        parentIndexNode.hierarchy_parents, levelList)

                        listOfLevels = [
                            {"field": itemDim, "name": item["name"]}]
                        indexParents = levelNode.hierarchy_parents
                        buildLevels(indexParents, listOfLevels)
                        item["levels"] = listOfLevels
                elif "datetime" in result.coords[itemDim].dtype.name:
                    item["numberFormat"] = "2,DD,0,,0,0,4,0,$,5,FULL,0"

                res.append(item)
        return res

    def isIndexed(self, result):
        if not result is None:
            obj = result
            if isinstance(obj, pd.Series):
                obj = pd.DataFrame({"values": obj})
            if isinstance(obj, pd.DataFrame):
                if (isinstance(obj.index, pd.MultiIndex) or isinstance(obj.index, pd.Index)) and len(obj.index.names) > 0 and (not obj.index.names[0] is None):
                    return True
        return False

    def getIndexValues(self, nodeDic, data: IndexValuesReq, result=None):
        res = []
        if data.node_id:
            if (not data.node_id is None) & (data.node_id in nodeDic):
                node = nodeDic[data.node_id]
                if result is None:
                    result = node.result
                res = self.checkDateFormat(
                    result[data.index_id].values).tolist()
        elif (not data.index_id is None) & (data.index_id in nodeDic):
            node = nodeDic[data.index_id]
            if result is None:
                result = node.result
            elif isinstance(result, np.ndarray):
                res = self.checkDateFormat(result).tolist()
            else:
                res = list(result)
        if data.text1:
            text1 = data.text1.lower()
            if data.filter == filterChoices.CONTAINS.value:
                res = list(
                    filter(lambda item: text1 in str(item).lower(), res))
            elif data.filter == filterChoices.NOT_CONTAINS.value:
                res = list(
                    filter(lambda item: not text1 in str(item).lower(), res))
        return res

    def getIndexType(self, nodeDic, nodeId, indexId):
        numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
        res = "S"

        if (not indexId is None) & (indexId in nodeDic):
            node = nodeDic[indexId]
            if isinstance(node.result, pd.Index):
                if str(node.result.values.dtype) in numerics:
                    res = "N"
            elif isinstance(node.result, np.ndarray):
                if str(node.result.dtype) in numerics:
                    res = "N"
        elif nodeId:
            if (not nodeId is None) & (nodeId in nodeDic):
                node = nodeDic[nodeId]
                if str(node.result.coords[indexId].values.dtype) in numerics:
                    res = "N"
        return res

    def getCubeValues(self, result, nodeDic, nodeId, query):
        if isinstance(result, xr.DataArray):
            res = {
                "dims": [],
                "values": []
            }

            # fix field.array
            query["columns"] = [xx.split(".")[0] for xx in query["columns"]]

            _filters = {}
            if not query["filters"] is None:
                for dimFilter in query["filters"]:
                    field = str(dimFilter["field"]).split(".")[0]
                    if self.hasDim(result, field):
                        dimFilter["values"] = [{"value": xx}
                                               for xx in dimFilter["values"]]
                        self.addToFilter(nodeDic, dimFilter, _filters)

            _filteredResult = result
            if len(_filters):
                _filteredResult = result.sel(_filters)
            nodeIndexes = self.getIndexes(nodeDic[nodeId], result)
            nodeIndexes = [xx.split(".")[0] for xx in nodeIndexes]
            for col in query["columns"]:
                if col in nodeIndexes:
                    item = {
                        "field": col,
                        "count": 0,
                        "values": [str(v) for v in self.checkDateFormat(_filteredResult.coords[col].values).tolist()]
                        # "values": result.filter(_filters).axis(col).values.tolist()
                    }
                    item["count"] = len(item["values"])
                    res["dims"].append(item)

            otherDims = [
                xx for xx in _filteredResult.dims if xx not in query["columns"]]

            if len(otherDims) > 0:
                squeezable = [_filteredResult.sizes[xx]
                              == 1 for xx in otherDims]
                to_squeeze = list(np.array(otherDims)[squeezable])
                if len(to_squeeze) > 0:
                    _filteredResult = _filteredResult.squeeze(to_squeeze)
                    otherDims = [
                        xx for xx in _filteredResult.dims if xx not in query["columns"]]

            resultValues = None
            if len(otherDims) > 0:
                resultValues = _filteredResult.sum(otherDims)
            else:
                resultValues = _filteredResult

            if isinstance(resultValues, xr.DataArray):
                # chec if haver nan values
                if pd.isnull(resultValues).any():
                    new_values = None
                    try:
                        new_values = np.where(
                            np.isnan(resultValues), None, resultValues)
                    except:
                        new_values = resultValues.values
                        new_values[pd.isnull(new_values)] = None
                    resultValues = resultValues.copy(data=new_values)

                if len(query["columns"]) > 0:
                    res["values"] = resultValues.transpose(
                        *query["columns"]).values.reshape(resultValues.size).tolist()
                else:
                    res["values"] = [resultValues.values.tolist()]
            else:
                res["values"] = resultValues
            return res

    def getCubeDimensionValues(self, result, nodeDic, nodeId, query):
        if isinstance(result, xr.DataArray):
            if len(query["columns"]) > 0:
                dimension = query["columns"][-1]

            if (dimension+"."+nodeId) in self.getIndexes(nodeDic[nodeId], result):
                finalList = [
                    str(v) for v in self.checkDateFormat(result.coords[dimension].values).tolist()[:1000]]
                finalList.sort()
                return finalList
        return []

    def getCubeMetadata(self, result, nodeDic, nodeId):
        res = None
        if isinstance(result, xr.DataArray):
            res = {
                "dims": [],
                "measures": [],
                "aggregator": "sum",
                "isEditable": True if self.isTable(nodeDic[nodeId]) == "1" else False,
                "nodeProperties": {
                    "title": nodeDic[nodeId].title if not nodeDic[nodeId].title is None else nodeDic[nodeId].identifier,
                    "numberFormat": nodeDic[nodeId].numberFormat,
                    "resultType": json.dumps(self.getStructure(result))
                }
            }

            # check if is in scenario
            if nodeDic[nodeId].model.isNodeInScenario(nodeDic[nodeId].identifier):
                res["nodeProperties"]["scenario"] = True

            for dim in result.dims:
                indexPart = str(dim).split(".")[0]
                itemDim = {
                    "field": dim,
                    "name": indexPart
                }
                if indexPart in nodeDic:
                    if not nodeDic[indexPart].title is None:
                        itemDim["name"] = nodeDic[indexPart].title
                    if nodeDic[indexPart].numberFormat:
                        itemDim["numberFormat"] = nodeDic[indexPart].numberFormat

                res["dims"].append(itemDim)

            res["measures"].append({
                "field": "datavalue",
                "name": "datavalue"
            })

        return res

    def isTable(self, node):
        if isinstance(node.result, xr.DataArray) and node.nodeClass == "inputtable":
            return "1"
        return "0"

    def setNodeValueChanges(self, nodeDic, nodeId, nodeChanges):
        node = nodeDic[nodeId]
        if self.isTable(node):
            for change in nodeChanges["changes"]:
                newValue = change["definition"]
                filters = {}
                for filterItem in change["filterList"]:
                    aux = {
                        "field": filterItem["Key"],
                        "values": [{
                            "value": filterItem["Value"]
                        }]}
                    self.addToFilter(nodeDic, aux, filters)

                for key in filters:
                    filters[key] = slice(filters[key][0], filters[key][0])

                node.result.loc[filters] = newValue

            node.definition = self.generateNodeDefinition(nodeDic, nodeId)
            return "ok"

    def generateNodeDefinition(self, nodeDic, nodeId, forceXArray=False, **kargs):
        node = nodeDic[nodeId]
        da = node.result
        """Generate code for cube definition"""
        np.set_printoptions(threshold=np.prod(da.values.shape))
        data = np.array2string(da.values, separator=",", precision=20, formatter={
                               'float_kind': lambda x: "np.nan" if np.isnan(x) else repr(x)}).replace('\n', '')

        _input_properties = None
        if "_input_properties" in node.definition:
            _input_properties = node.model.evaluate(node.definition.split("# values")[
                                                    0] + "result = _input_properties")
        else:
            if self.kindToString(da.values.dtype.kind) == "string" or self.kindToString(da.values.dtype.kind) == "object":
                _input_properties = {"defaultValue": ""}
            elif np.issubdtype(da.values.dtype, np.integer):
                _input_properties = {"defaultValue": 0}
            else:
                _input_properties = {"defaultValue": 0.}

        if "defaultValue" in kargs:
            _input_properties["defaultValue"] = kargs["defaultValue"]

        coords = []
        for dim in list(da.dims):
            coord_data = json.dumps(da[dim].values.tolist())
            item_coord = f'({dim}.name, {coord_data})' if dim in nodeDic else f"({dim} ,{coord_data})"
            coords.append(item_coord)

        str_coords = f'[{",".join(coords)}]'
        data_deff = f'result = xr.DataArray({data}, coords={str_coords})'

        if isinstance(_input_properties["defaultValue"], str):
            data_deff += '.astype("O")'

        deff = f"# properties\n_input_properties = {_input_properties}\n\n# values\n{data_deff}"

        return deff

    def dumpNodeToFile(self, nodeDic, nodeId, fileName):
        definition = self.generateNodeDefinition(nodeDic, nodeId)
        with open(fileName, 'w') as f:
            f.write(definition)
            f.close()

    def applyHierarchy(self, result, nodeDic, nodeId, dims, rows, columns, sby):

        def hierarchize(dataArray, levels, maps, hierarchyDic):

            mapArray = nodeDic[maps[0]].result
            coordValues = mapArray.values.copy()
            targetIndexId = nodeDic[levels[1]].result.name

            for pos, level in enumerate(levels):
                if pos > 0:

                    if not maps[pos] is None:
                        mapArrayLevel = nodeDic[maps[pos]].result

                        for ii in range(len(coordValues)):
                            if not coordValues[ii] is None:
                                try:
                                    newVal = mapArrayLevel.sel(
                                        {mapArrayLevel.dims[0]: coordValues[ii]}, drop=True).values.item(0)
                                    coordValues[ii] = newVal
                                except Exception as ex:
                                    coordValues[ii] = None
                                    #raise ValueError("Hierarchy not found. Level: " + targetIndexId + ", value: " + coordValues[ii])
                                    pass

            # perform aggregate
            dataArray.coords[levels[0]].values = coordValues
            _df = dataArray.to_series()
            _df = _df.groupby(list(dataArray.dims), sort=False).agg(sby)
            _da = _df.to_xarray()
            reindex_dic = dict()
            for dimension in _da.dims:
                if dimension == levels[0]:
                    reindex_dic[dimension] = nodeDic[levels[-1:]
                                                     [0]].result.values
                elif dimension in nodeDic and isinstance(nodeDic[dimension].result, pd.Index):
                    node_id = dimension
                    if not hierarchyDic is None and dimension in hierarchyDic:
                        node_id = hierarchyDic[dimension]

                    reindex_dic[dimension] = nodeDic[node_id].result.values

            _db = _da.reindex(reindex_dic)

            return _db

        allDims = (dims or []) + (rows or []) + (columns or [])

        hierarchyDic = dict()
        for dim in allDims:
            if dim and dim["currentLevel"] and dim["currentLevel"] != str(dim["field"]).split(".")[0]:
                hierarchyDic[str(dim["field"]).split(".")[
                    0]] = dim["currentLevel"]

                # recursive fn for search parent

                def findPath(indexNode, level, levels, maps):
                    if indexNode.identifier == level:
                        levels.append(indexNode.identifier)
                        maps.append(None)
                        return True
                    else:
                        _for_calc = indexNode.result
                        parents = indexNode.hierarchy_parents
                        if parents is None:
                            return False
                        else:
                            if not isinstance(parents, list):
                                parents = [parents]
                            mapArrays = indexNode.hierarchy_maps
                            if not isinstance(mapArrays, list):
                                mapArrays = [mapArrays]

                            mapPos = 0
                            for parentId in parents:
                                parent = nodeDic[parentId]
                                if findPath(parent, level, levels, maps):
                                    levels.append(indexNode.identifier)
                                    maps.append(mapArrays[mapPos])
                                    return True
                                mapPos += 1
                    return False

                field = str(dim["field"]).split(".")[0]
                currentLevel = dim["currentLevel"]
                indexNode = nodeDic[field]

                levels = []
                maps = []
                findPath(indexNode, currentLevel, levels, maps)
                levels.reverse()
                maps.reverse()

                result = hierarchize(result.copy(), levels, maps, hierarchyDic)
        return result

    def geoUnclusterData(self, result, nodeDic, nodeId, rowIndex, attIndex, latField="latitude", lngField="longitude", geoField="geoField", labelField="labelField", sizeField="sizeField", colorField="colorField", iconField="iconField"):
        _tmp_for_geo = pd.Index([latField, lngField, geoField, labelField,
                                 sizeField, colorField, iconField], name="tmp_for_geo")
        attIndex = attIndex.split(".")[0]
        rowIndex = rowIndex.split(".")[0]
        _idx = nodeDic[attIndex].result
        rowIndexObj = nodeDic[rowIndex].result

        _temp = result.reindex({_idx.name: _tmp_for_geo.values})
        _temp[_tmp_for_geo.name] = _temp[_idx.name]
        _temp = _temp.swap_dims({_idx.name: _tmp_for_geo.name}).drop(_idx.name)

        mapCube = _temp.transpose(*[rowIndex, "tmp_for_geo"]).values
        res = dict()
        points = []
        pos = 0
        for itemRow in mapCube:
            vo = dict()
            vo["id"] = str(rowIndexObj.values[pos])
            vo["lat"] = itemRow[0]
            vo["lng"] = itemRow[1]
            vo["geoDef"] = itemRow[2]
            vo["labelRes"] = itemRow[3]
            vo["sizeRes"] = itemRow[4]
            vo["colorRes"] = itemRow[5]
            vo["iconRes"] = itemRow[6]
            points.append(vo)
            pos += 1
        res["points"] = points

        for nn, point in enumerate(res["points"]):
            if nn == 0:
                try:
                    if not math.isnan(float(point["sizeRes"])):
                        res["minSize"] = float(point["sizeRes"])
                        res["maxSize"] = float(point["sizeRes"])
                except Exception as ex:
                    pass

                try:
                    if not math.isnan(float(point["colorRes"])):
                        res["minColor"] = float(point["colorRes"])
                        res["maxColor"] = float(point["colorRes"])
                except Exception as ex:
                    pass

                try:
                    if not math.isnan(float(point["iconRes"])):
                        res["minIcon"] = float(point["iconRes"])
                        res["maxIcon"] = float(point["iconRes"])
                except Exception as ex:
                    pass
            else:
                try:
                    if not math.isnan(float(point["sizeRes"])):
                        if point["sizeRes"] > res["maxSize"]:
                            res["maxSize"] = point["sizeRes"]
                        if point["sizeRes"] < res["minSize"]:
                            res["minSize"] = point["sizeRes"]
                except Exception as ex:
                    pass

                try:
                    if not math.isnan(float(point["colorRes"])):
                        if point["colorRes"] > res["maxColor"]:
                            res["maxColor"] = point["colorRes"]
                        if point["colorRes"] < res["minColor"]:
                            res["minColor"] = point["colorRes"]
                except Exception as ex:
                    pass

                try:
                    if not math.isnan(float(point["iconRes"])):
                        if point["iconRes"] > res["maxIcon"]:
                            res["maxIcon"] = point["iconRes"]
                        if point["iconRes"] < res["minIcon"]:
                            res["minIcon"] = point["iconRes"]
                except Exception as ex:
                    pass
        return res

    def postCalculate(self, node, result):
        """Method executed after calculate node
        """
        if isinstance(result, xr.DataArray):
            result.name = node.title

    def copyAsValues(self, result, nodeDic, nodeId):
        """ Copy node as values """
        node = nodeDic[nodeId]
        newDef = ""
        if isinstance(result, float) or isinstance(result, int):
            newDef = "result = " + str(result)
        elif isinstance(result, xr.DataArray):
            newDef = self.generateNodeDefinition(nodeDic, nodeId, True)
        else:
            return False

        # try new definition
        temp_res = node.model.evaluate(newDef)

        node.nodeClass = "inputtable"
        nodeFormat = node.model.getDefaultNodeFormat(node.nodeClass)
        node.fromObj(nodeFormat)
        node.definition = newDef

        return True

    def previewNode(self, nodeDic, nodeId):
        if not nodeDic[nodeId].result is None:
            _result = nodeDic[nodeId].result
            res = {
                "resultType": str(type(_result)),
                "dims": [],
                "console": nodeDic[nodeId].lastEvaluationConsole,
                "preview": ""
            }
            _preview = ""
            _preview = f"{type(_result)} - dtype: {(_result.dtype)}"
            _preview += f"\n\n{_result.coords}"
            _preview += f"\n\nTotal cells: {_result.sizes} = {_result.size} cells"
            _preview += f"\n\nSample data: \n {_result.data}"
            res["preview"] = _preview

            return json.dumps(res)
        else:
            return self.generateEmptyPreviewResponse(nodeDic, nodeId)

    def kindToString(self, kind):
        """Returns the data type on human-readable string
        """
        if kind in {'U', 'S'}:
            return "string"
        elif kind in {'b'}:
            return "boolean"
        elif kind in {'i', 'u', 'f', 'c'}:
            return "numeric"
        elif kind in {'m', 'M'}:
            return "date"
        elif kind in {'O'}:
            return "object"
        elif kind in {'V'}:
            return "void"



import datetime as dt
import io
import re
import uuid
from contextlib import redirect_stdout
from sys import exc_info, getsizeof
from types import CodeType

import numpy as np
import pandas as pd
import xarray as xr

from pyplan_core.classes.evaluators.Evaluator import Evaluator
from pyplan_core.classes.dynamics.BaseDynamic import BaseDynamic
from pyplan_core.classes.dynamics.FactoryDynamic import FactoryDynamic
from pyplan_core.cubepy.Helpers import Helpers
from pyplan_core.classes.IOEngine import IOEngine
from pyplan_core.classes.NodeInfo import NodeInfo
from pyplan_core.classes.PyplanFunctions import Selector


class BaseNode(object):

    SERIALIZABLE_PROPERTIES = ['identifier', 'definition', 'title', 'nodeClass', 'moduleId', 'x', 'y', 'z', 'w', 'h',
                               "description", "units", "color", "errorInDef", "nodeInfo", "nodeFont", "numberFormat", "originalId", "extraData", "picture", "evaluateOnStart"]

    FORMNODE_TYPE_CHECKBOX = 0
    FORMNODE_TYPE_COMBOBOX = 1
    FORMNODE_TYPE_SCALAR = 2
    FORMNODE_TYPE_BUTTON = 3

    def __init__(self, model, identifier=None, nodeClass=None, moduleId=None, x=None, y=None, originalId=None):
        self._model = model
        self._originalId = originalId
        self._result = None
        self._isCalc = False
        self._title = None
        self._description = None
        self.units = None
        self.numberFormat = None
        self.color = None
        self.errorInDef = False
        self.nodeClass = nodeClass if not nodeClass is None else "variable"
        self._definition = self.getDefaultDefinitionByClass(self.nodeClass)
        self.nodeInfo = self.getDefaultNodeinfoByClass(self.nodeClass)
        self.moduleId = moduleId if not moduleId is None else self._model.modelNode.identifier
        self.x = int(x) if not x is None else 100
        self.y = int(y) if not y is None else 100
        self.z = 1
        if nodeClass == "text":
            self.z = -1
        self.w = 100
        self.h = 50

        self._ioEngine = IOEngine(self)
        self._resultMemory = 0
        self.system = False
        self.lastEvaluationTime = 0
        self.lastEvaluationConsole = ""
        self.lastLazyTime = 0
        self.evaluationVersion = 0
        self.profileParent = None
        self.nodeFont = None
        self._bypassCircularEvaluator = False
        self.extraData = None
        self.dynamicEvaluator = None
        self.picture = None
        self._hierarchy_parents = None
        self._hierarchy_maps = None
        self._releaseMemory = False
        self._evaluateOnStart = False

        # load default props by toolbar
        nodeFormat = model.getDefaultNodeFormat(self.nodeClass)

        if identifier is None:
            if not nodeClass is None:
                if nodeClass == "alias" or nodeClass == "formnode" or nodeClass == "text":
                    self._identifier = "a"+str(uuid.uuid4().hex).lower()
                elif nodeClass == "index":
                    self._identifier = model.getNextIdentifier("index")
                else:
                    self._identifier = model.getNextIdentifier("node")
            if not nodeFormat is None:
                if not nodeFormat["nodeClass"] is None:
                    if nodeFormat["nodeClass"] == "alias" or nodeFormat["nodeClass"] == "formnode" or nodeFormat["nodeClass"] == "text":
                        self._identifier = "a"+str(uuid.uuid4().hex).lower()
                    elif nodeFormat["nodeClass"] == "index":
                        self._identifier = model.getNextIdentifier("index")
                    else:
                        self._identifier = model.getNextIdentifier("node")
        else:
            self._identifier = identifier

        if not nodeFormat is None:
            self.fromObj(nodeFormat)

    def __del__(self):
        self.release()

    def release(self):
        """Release object"""
        self._model = None
        self._result = None
        self._identifier = None
        self._definition = None
        self.moduleId = None
        self.dynamicEvaluator = None
        if not self.ioEngine is None:
            self.ioEngine.release()
            self._ioEngine = None

    def toObj(self, exceptions=None, properties=None, fillDefaultProperties=False):
        """Convert node to dictionary"""
        res = {}
        if not properties is None:
            for k in properties:
                if k in BaseNode.SERIALIZABLE_PROPERTIES and hasattr(self, k):
                    res[k] = getattr(self, k)
        else:
            for k in BaseNode.SERIALIZABLE_PROPERTIES:
                if ((exceptions is None) or (not k in exceptions)) and hasattr(self, k):
                    res[k] = getattr(self, k)

        if fillDefaultProperties:
            if not "color" in res or res["color"] is None:
                res["color"] = BaseNode.getDefaultColor(self.nodeClass)

            res["hasPicture"] = self.hasPicture

            # fill formnode/alias
            if self.nodeClass in ["formnode", "alias"]:
                res["originalId"] = self.originalId
                res["originalClass"] = self.originalClass

                if self.nodeClass == "formnode":
                    res["formNodeType"] = self.formNodeType
                    res["formNodeExtraValue"] = self.formNodeExtraValue

                    # check if value is serializable
                    try:
                        # jsonify(self.formNodeValue)
                        # comentado porque no se puede utilizar jsonify fuera del contexto del pyplan_engine. (cuando se usa multiprocesador)
                        res["formNodeValue"] = self.formNodeValue
                    except TypeError as ex:
                        res["formNodeValue"] = str(ex)
                        res["formNodeType"] = self.FORMNODE_TYPE_SCALAR

        return res

    def fromObj(self, obj):
        """Convert dictionary to basenode object"""
        for k in BaseNode.SERIALIZABLE_PROPERTIES:
            if k in obj:
                # work around for migrate nodeInfo class
                if k == "nodeInfo":
                    if isinstance(obj[k], dict):
                        obj[k] = NodeInfo(obj[k]["showInputs"], obj[k]["showOutputs"], obj[k]
                                          ["showLabel"], obj[k]["showBorder"], obj[k]["fill"], obj[k]["useNodeFont"])

                setattr(self, k, obj[k])

    # Props

    @property
    def node(self):
        return self

    @property
    def model(self):
        return self._model

    @property
    def result(self):
        try:
            self.calculate()
        except Exception as e:
            err = str(e)
            if not "Error evaluating" in err:
                title = self.title
                if title is None:
                    title = ""
                else:
                    title = title+" "
                title = title + "(" + self.identifier + ") "
                err = "Error evaluating "+title+": " + err

            raise ValueError(err)

        try:
            return self._result
        finally:
            if self.releaseMemory:
                self._isCalc = False
                self.profileParent = None
                self._result = None
                self._resultMemory = 0

    @property
    def isCalc(self):
        return self._isCalc

    @property
    def releaseMemory(self):
        return self._releaseMemory

    @releaseMemory.setter
    def releaseMemory(self, value):
        self._releaseMemory = value

    @property
    def evaluateOnStart(self):
        return self._evaluateOnStart

    @evaluateOnStart.setter
    def evaluateOnStart(self, value):
        self._evaluateOnStart = value

    @property
    def resultType(self):
        if self._isCalc:
            return str(type(self.result))
        else:
            return ""

    @property
    def usedMemory(self):
        return (getsizeof(self.definition) + self._resultMemory) / 1024 / 1024

    @property
    def hasPicture(self):
        if self.picture:
            return True
        return False

    @property
    def title(self):
        if not self.originalId is None:
            if self.model.existNode(self.originalId):
                return self.model.getNode(self.originalId).title
            else:
                return self._title
        else:
            return self._title

    @title.setter
    def title(self, value):
        self._title = value
        if self.isCalc:
            self.postCalculate()

    @property
    def description(self):
        if not self.originalId is None and self.model.existNode(self.originalId):
            return self.model.getNode(self.originalId).description
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, value):
        if not self.model.isLoadingModel:
            value = self.sanitizeDefinition(value)

        self._definition = value
        if not self.model.isLoadingModel:
            self.invalidate()
            self.generateIO()

        self._isCalc = False

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, value):
        if value != self._identifier:
            if self._model.existNode(value):
                raise ValueError("'The id '" + value + "' already exists")
            if self._model.updateNodeIdInDic(self._identifier, value):
                self.ioEngine.updateNodeId(self._identifier, value)
                self._identifier = value
                if self.isCalc:
                    self.invalidate()

    @property
    def isIndexed(self):
        raise ValueError("TODO: Mover isIndexed a clase BaseEvaluator")
        return False

    @property
    def indexes(self):
        res = []
        self.calculate()
        if not self._result is None:
            evaluator = Evaluator.createInstance(self._result)
            return evaluator.getIndexes(self)

    @property
    def ioEngine(self):
        return self._ioEngine

        """
        #comentado porque da error al eliminar alias y dps al grabar modelo
        if self.originalId is not None:            
            return self.model.getNode(self.originalId)._ioEngine
        else:
            return self._ioEngine
        """

    @property
    def inputs(self):
        return self.ioEngine.inputs

    @property
    def outputs(self):
        return self.ioEngine.outputs

    @property
    def originalId(self):
        return self._originalId

    @originalId.setter
    def originalId(self, value):
        self._originalId = value

    @property
    def originalClass(self):
        if self._originalId is not None:
            if self.model.existNode(self._originalId):
                return self.model.getNode(self._originalId).nodeClass
        return None

    @property
    def formNodeType(self):
        if not self.originalId is None:
            try:
                originalNode = self.model.getNode(self.originalId)
                if not originalNode is None:
                    if "choice(" in originalNode.definition or "selector(" in originalNode.definition:
                        return self.FORMNODE_TYPE_COMBOBOX
                    elif str(originalNode.result).isnumeric() or isinstance(originalNode.result, str) or isinstance(originalNode.result, float) or isinstance(originalNode.result, int):
                        return self.FORMNODE_TYPE_SCALAR
                    elif isinstance(originalNode.result, bool):
                        return self.FORMNODE_TYPE_CHECKBOX

            except (RuntimeError, TypeError, NameError, ValueError, SyntaxError):
                pass

        return self.FORMNODE_TYPE_BUTTON

    @property
    def formNodeValue(self):
        if self.formNodeType == self.FORMNODE_TYPE_SCALAR or self.formNodeType == self.FORMNODE_TYPE_CHECKBOX or self.formNodeType == self.FORMNODE_TYPE_COMBOBOX:
            try:
                res = self.model.getNode(self.originalId).result
                if isinstance(res, Selector):
                    if "numpy." in str(type(res.value)):
                        return res.value.item()
                    else:
                        return res.value
                else:
                    return res
            except Exception as e:
                self.errorInDef = True
        return None

    @property
    def formNodeExtraValue(self):
        if self.formNodeType == self.FORMNODE_TYPE_SCALAR or self.formNodeType == self.FORMNODE_TYPE_CHECKBOX:
            try:
                if isinstance(self.result, str):
                    return f'"{self.result}"'
                else:
                    return self.result
            except (RuntimeError, TypeError, NameError, ValueError, SyntaxError):
                pass
        elif self.formNodeType == self.FORMNODE_TYPE_COMBOBOX:
            originalNode = self.model.getNode(self.originalId)
            return originalNode.definition
        return None

    @property
    def hierarchy_parents(self):
        return self._hierarchy_parents

    @property
    def hierarchy_maps(self):
        return self._hierarchy_maps

    # Synonimos
    @property
    def isin(self):
        return self.moduleId

    # Methods

    def preDelete(self):
        self.invalidate()
        self.ioEngine.updateOnDeleteNode()

    def silentInvalidate(self):
        self._isCalc = False
        self.profileParent = None
        self._result = None
        self._resultMemory = 0

    def invalidate(self, fromCircularNode=False):
        """Invalidate node result"""
        self._isCalc = False
        self.profileParent = None
        self._result = None
        self._resultMemory = 0
        if self.isCircular() and not fromCircularNode:
            circularNodes = self.getSortedCyclicDependencies()
            for node in circularNodes:
                if node != self.identifier:
                    if self._model.existNode(node):
                        self._model.getNode(node).invalidate(
                            fromCircularNode=True)
        self.invalidateOutputs()

    def invalidateOutputs(self):
        """Invalidate outputs of this node"""
        isCircular = self.isCircular()
        for node in self.ioEngine.outputs:
            if not node is None:
                if self._model.existNode(node):
                    if isCircular and self._model.getNode(node).isCircular() and self.identifier in self._model.getNode(node).getSortedCyclicDependencies():
                        pass
                    else:
                        if self._model.getNode(node).isCalc:
                            self._model.getNode(node).invalidate()

    def generateIO(self):
        """Generate inputs and outputs"""
        if self.nodeClass not in ['alias', 'formnode']:
            tmpCode = None
            try:
                tmpCode = self.compileDef(self._definition)
            except SyntaxError:
                aa = 1
                # TODO: get detail of error

            if not tmpCode is None:
                # TODO: if cp.excel.. in def and other node Id is excel then don't generate input for that node
                nameList = self.parseNames(tmpCode)
                finalList = []
                for item in nameList:
                    if item and self.model.existNode(item):
                        rxText = "[^!.a-zA-Z_]\\b(###)\\b(?!={1}(?!={1}))".replace(
                            "###", item)
                        rx = re.compile(rxText)
                        if len(re.findall(rx, self.definition)) > 0:
                            finalList.append(item)
                self.ioEngine.updateInputs(finalList)

    def compileDef(self, definition):
        """Compile definition"""
        try:
            self.errorInDef = False
            tmpCode = compile(str(definition), '<string>', 'exec')
            return tmpCode
        except SyntaxError:
            self.errorInDef = True
            # TODO: get detail of error
            pass
        return None

    def _getCalcNode(self, node):
        lazy_start_time = dt.datetime.now()
        try:
            return self._model.getNode(node).result
        finally:
            lazy_end_time = dt.datetime.now()
            if self.identifier != node:
                self.lastLazyTime = self.lastLazyTime + \
                    (lazy_end_time - lazy_start_time).total_seconds()

    def calculate(self, extraParams=None):
        """Calculate result of the node"""

        if not self.isCalc or self.nodeClass == "button" or self.nodeClass == "formnode":
            nodeIsCircular = self.isCircular()
            if not self._bypassCircularEvaluator and nodeIsCircular:
                circularNodes = self.getSortedCyclicDependencies()

                if self.dynamicEvaluator is None:
                    self.dynamicEvaluator = FactoryDynamic.createInstance(
                        circularNodes, self)

                params = self.dynamicEvaluator.generateCircularParameters(
                    self, circularNodes)

                if params['dynamicIndex'] is None:
                    raise ValueError("Cyclic dependency detected between nodes: " + ",".join(
                        circularNodes) + ".\nPlease use the 'dynamic' function")
                self.dynamicEvaluator.circularEval(self, params)
            else:
                from_circular_evaluator = self._bypassCircularEvaluator

                self.sendStartCalcNode(from_circular_evaluator)
                self.model.currentProcessingNode(self.identifier)
                self._bypassCircularEvaluator = False

                startTime = dt.datetime.now()
                finalDef = str(self._definition)
                self.lastLazyTime = 0

                # CLEAR circular dependency
                if nodeIsCircular:
                    finalDef = BaseDynamic.clearAllCircularDependency(finalDef)
                tmpCode = self.compileDef(finalDef)

                # check for replace calls to varaibles with next rules:
                # node1  --> change to getNode('node1').result
                # node1.result  --> change to getNode('node1').result
                # node1.title   --> change to getNode('node1').title
                # "node1" --> no change
                # 'node1' --> no change
                if not tmpCode is None:
                    names = self.parseNames(tmpCode)
                    rx = r"('[^'\\]*(?:\\.[^'\\]*)*'|\"[^\"\\]*(?:\\.[^\"\\]*)*\")|\b{0}\b"
                    for node in names:
                        if self._model.existNode(self._model.clearId(node)):
                            finalDef = re.sub(rx.format(node), lambda m:
                                              (
                                m.group(1)
                                if m.group(1)
                                else
                                (
                                    "getNode('"+node+"')"
                                    if (m.endpos > m.regs[0][1]+5) and ((m.string[m.regs[0][1]:m.regs[0][1]+5] == '.node') or (m.string[m.regs[0][1]:m.regs[0][1]+8] == '.timeit('))
                                    else
                                    (node
                                     if (m.string[m.regs[0][0]-1:m.regs[0][0]+len(node)] == ('.'+node))
                                     else "getCalcNode('"+node+"')"
                                     )
                                )
                            ), finalDef, 0, re.IGNORECASE)
                        elif node == "self":
                            finalDef = re.sub(rx.format(node), lambda m:
                                              (
                                m.group(1)
                                if m.group(1)
                                else
                                "getNode('" + self.identifier + "')"
                                if (m.endpos > m.regs[0][1]+11) and (m.string[m.regs[0][1]:m.regs[0][1]+11] != '._tryFilter')
                                else "self"
                            ), finalDef, 0, re.IGNORECASE)

                localRes = {
                    "getNode": self._model.getNode,
                    "getCalcNode": self._getCalcNode,
                    "cp": Helpers(self)
                }
                if not extraParams is None:
                    for keyParam in extraParams:
                        localRes[keyParam] = extraParams[keyParam]
                
                customImports = self.model.getCustomImports()
                if customImports:
                    for keyParam in customImports:
                        localRes[keyParam] = customImports[keyParam]

                try:
                    # execute node definition in supervised context
                    memoryIO = io.StringIO()
                    try:
                        with redirect_stdout(memoryIO):
                            exec(compile(finalDef, '<string>', 'exec'), localRes)
                    except Exception as ex:
                        if "_io.StringIO" in str(ex):
                            exec(compile(finalDef, '<string>', 'exec'), localRes)
                        else:
                            raise ex

                    self.lastEvaluationConsole = memoryIO.getvalue()
                    memoryIO = None

                    if self.nodeClass not in ["button", "module", "text"]:
                        if 'this' in localRes:
                            self._result = localRes['this']
                            self._resultMemory = Helpers.getResultSize(
                                self._result)
                        elif 'result' in localRes:
                            self._result = localRes['result']
                            self._resultMemory = Helpers.getResultSize(
                                self._result)
                        else:
                            self._result = None
                            if self.lastEvaluationConsole != "":
                                self._result = str(self.lastEvaluationConsole)
                            else:
                                raise ValueError(
                                    "The result was not found. Did you forget to include the text 'result =' ?")

                    self._isCalc = self.nodeClass != "button"
                    self.postCalculate()

                    endTime = dt.datetime.now()
                    self.lastEvaluationTime = (
                        endTime - startTime).total_seconds() - self.lastLazyTime
                    if self.lastEvaluationTime < 0:
                        self.lastEvaluationTime = 0
                    self.evaluationVersion = self.model.evaluationVersion
                finally:
                    localRes["cp"].release()
                    localRes = None
                    self.sendEndCalcNode(from_circular_evaluator)
        else:
            self._bypassCircularEvaluator = False

    def postCalculate(self):
        evaluator = Evaluator.createInstance(self._result)
        evaluator.postCalculate(self, self._result)
        if self.nodeClass=="inputtable":
            self.validateInputTable(evaluator)

    def validateInputTable(self,evaluator):
        """Validate inputs of the inputtable node"""
        da = self._result
        if isinstance(da,xr.DataArray):
            has_changed=False
            for dim in da.dims:
                node_index = self.model.getNode(dim)
                if list(node_index.result.values) != list(da.coords[dim].values): #values of index has changed
                    has_changed = True
                    _input_properties=None
                    if "_input_properties" in self.definition:
                        _input_properties =  self.model.evaluate(self._definition.split("# values")[0] + "result = _input_properties")
                    else: 
                        _input_properties = {"defaultValue":0.}
                    da = da.reindex({dim: list(node_index.result.values)})
                    da = da.fillna(_input_properties["defaultValue"])

            if has_changed:
                self._result = da
                self._definition = evaluator.generateNodeDefinition(self.model.nodeDic, self.identifier)


    def parseNames(self, compiledCode):
        """Parse names used in node definition"""
        res = []
        if not compiledCode is None:
            res = compiledCode.co_names
            for co in compiledCode.co_consts:
                if not co is None and isinstance(co, CodeType):
                    res += co.co_names
        return res

    def updateDefinitionForChangeId(self, oldId, newId):
        """ Search and replace on this node definition the oldId for the newId
        """
        rx = re.compile(r"((?<!\w)(?<!\d)(?<!\.)" + re.escape(oldId) +
                        "(?!\d)(?!\w))(?=(?:[^\"]|[\"][^\"]*[\"])*$)")
        newDef = re.sub(rx, newId, self.definition)
        self.definition = newDef
        pass

    def isCircular(self):
        """ Checks if the node is part of a cycle
        """
        _id = self.identifier if self.originalId is None else self.originalId
        _inputNodes = [_id]

        nn = 0
        while nn < len(_inputNodes):
            _node = _inputNodes[nn]
            if self.model.existNode(_node):
                for _inputId in self.model.getNode(_node).ioEngine.inputs:
                    if _inputId == _id:
                        return True
                    else:
                        if not _inputId in _inputNodes:
                            _inputNodes.append(_inputId)
            nn += 1
        return False

    def getFullInputs(self):
        """
        Return list of all node inputs and inputs of inputs
        """
        res = [self.identifier if self.originalId is None else self.originalId]
        nn = 0
        while nn < len(res):
            _node = res[nn]
            if self.model.existNode(_node) and self.model.getNode(_node).ioEngine.inputs:
                for _inputId in self.model.getNode(_node).ioEngine.inputs:
                    input_node = self.model.getNode(_inputId)
                    if not _inputId in res:
                        res.append(_inputId)
            nn += 1
        return res

    def getSortedCyclicDependencies(self):
        """
        Return list of nodes in circular dependencyes, sortered by execution order
        """
        res = []
        if self.isCircular():
            res = [self.identifier if self.originalId is None else self.originalId]
            # fill node inputs
            nn = 0
            while nn < len(res):
                _node = res[nn]
                for _inputId in self.model.getNode(_node).ioEngine.inputs:
                    input_node = self.model.getNode(_inputId)
                    if not _inputId in res and input_node.isCircular():
                        # check if node is in circle of _inputId
                        if _node in input_node.getFullInputs():
                            res.append(_inputId)
                nn += 1
        return res

    def profileNode(self, evaluated, response, evaluationVersion, profileParentId):
        """Perform node profile"""
        if not evaluated is None:
            if self.identifier in evaluated:
                return response
            else:
                evaluated.append(self.identifier)
                if self.evaluationVersion == evaluationVersion:
                    if self.profileParent is None:
                        self.profileParent = profileParentId
                    auxVal = {"nodeId": self.identifier, "title": self.identifier if self.model.getNode(self.identifier).title is None else self.model.getNode(self.identifier).title, "moduleId": self.model.getNode(
                        self.identifier).moduleId, "evaluationTime": self.model.getNode(self.identifier).lastEvaluationTime, "usedMemory": self.model.getNode(self.identifier).usedMemory, "calcTime": 0.00},
                    response.extend(auxVal)
                if not response is None:
                    for nodeInputId in self.inputs:
                        if nodeInputId not in evaluated:
                            response = self.model.getNode(nodeInputId).profileNode(
                                evaluated, response, evaluationVersion, self.identifier)
        return response

    def sanitizeDefinition(self, value):
        """
        Sanitize from current front (Ade to py)
        """
        if str(value).isnumeric():
            value = "result = " + str(value)
        return value

    def getDefaultDefinitionByClass(self, nodeClass):
        """Return default definition by node class"""
        if nodeClass == "module":
            return ""
        elif nodeClass == "index":
            return "result = cp.index(['Item 1', 'Item 2', 'Item 3'])"
        elif nodeClass == "button":
            return ""
        else:
            return "result = 0"

    def getDefaultNodeinfoByClass(self, nodeClass):
        """
            Default nodeinfo by class. 
            Positions are:
                showInputs = 0
                showOutputs = 1
                showLabel = 2
                showBorder = 3
                fill = 4
                useNodeFont = 5
        """
        if nodeClass == "text":
            return NodeInfo(0, 0, 1, 0, 0, 0)
        if nodeClass == "alias":
            return NodeInfo(0, 0, 1, 1, 1, 0)
        elif nodeClass == "index":
            return NodeInfo(0, 0, 1, 1, 1, 0)
        elif nodeClass == "function":
            return NodeInfo(0, 0, 1, 1, 1, 0)
        elif nodeClass == "button":
            return NodeInfo(0, 0, 1, 1, 1, 0)
        else:
            return NodeInfo(1, 1, 1, 1, 1, 0)

    def timeit(self, repeat=1):
        """ Get stats after evaluate "repeat" times this node
        """
        import timeit

        def _testNode_():
            self.invalidate()
            self.node.result

        _locals_ = {"_testNode_": _testNode_}
        _res = timeit.repeat("_testNode_() ", repeat=repeat,
                             number=1, globals=_locals_)
        return str({
            "average": float(str(round(np.mean(_res), 3))),
            "best": float(str(round(np.min(_res), 3))),
            "max": float(str(round(np.max(_res), 3))),
        })

    def set_hierarchy(self, parents, maps):
        self._hierarchy_parents = parents
        self._hierarchy_maps = maps

    def sendStartCalcNode(self, fromCircularEvaluator=False, fromDynamic=False):
        if self.model.debugMode and not self.identifier in ["__evalnode__", "dynamic"] and not fromCircularEvaluator and self._model.ws:
            self._model.ws.sendDebugInfo(
                self.identifier, self.title if self.title else "", "startCalc", fromDynamic=fromDynamic)

    def sendEndCalcNode(self, fromCircularEvaluator=False, fromDynamic=False):
        if self.model.debugMode and not self.identifier in ["__evalnode__", "dynamic"] and not fromCircularEvaluator and self._model.ws:
            resources = None
            try:
                resources = self.model.getSystemResources(onlyMemory=True)
            except:
                resources = {
                    "usedMemory": 0,
                    "totalMemory": 0
                }

            self._model.ws.sendDebugInfo(
                self.identifier, self.title if self.title else "", "endCalc", self.lastEvaluationTime, resources["usedMemory"], resources["totalMemory"], fromDynamic=fromDynamic)


    # ***********************************
    # *** CYCLICK EVALUATOR  METHODS  ***
    # ***********************************

    def bypassCircularEvaluator(self):
        """Mark node for bypass circular evaluator"""
        self._bypassCircularEvaluator = True
        return self

    # *********************************
    # *** CLASS METHODS             ***
    # *********************************

    @staticmethod
    def getDefaultColor(nodeClass):
        dic = {
            "variable": "#4CBCFF",
            "button": "#a6a6a6",
            "decision": "#4cffa6",
            "constant": "#ff794c",
            "module": "#4c83ff",
            "function": "#cc99ff",
            "chance": "#49E4F7",
            "determ": "#076CBC",
            "objective": "#ff4c97",
            "index": "#9999ff",
            "library": "#076CBC",
            "form": "#076CBC",
            "text": "#FEFEFE",
            "picture": "#FEFEFE",
            "formnode": "#076CBC",
            "constraint": "#f2e340",
            "inputtable": "#ffd700"
        }
        return dic[str(nodeClass).lower()] if str(nodeClass).lower() in dic else "#6699FF"




class IOEngine(object):

    def __init__(self, node):
        self.node = node
        self.inputs = []
        self.outputs = []

    def release(self):
        self.inputs = None
        self.outputs = None
        self.node = None

    def updateInputs(self, names):
        # remove prior outputs
        for inputNode in self.inputs:
            if not inputNode in names:
                if self.node.model.existNode(inputNode):
                    self.node.model.getNode(inputNode).ioEngine.removeOutput(
                        self.node.identifier)

        newInputs = []
        for nodeId in names:
            if self.node.model.existNode(nodeId):
                newInputs.append(nodeId)
                if not nodeId in self.inputs:
                    self.node.model.getNode(nodeId).ioEngine.addOutput(
                        self.node.identifier)

        self.inputs = newInputs

    def removeOutput(self, nodeId):
        if nodeId in self.outputs:
            self.outputs.remove(nodeId)

    def removeInput(self, nodeId):
        if nodeId in self.inputs:
            self.inputs.remove(nodeId)

    def addOutput(self, nodeId):
        self.outputs.append(nodeId)

    def updateNodeId(self, oldId, newId):
        for inputNode in self.inputs:
            if self.node.model.existNode(inputNode):
                self.node.model.getNode(
                    inputNode).ioEngine.updateOutputId(oldId, newId)

        for outputNode in self.outputs:
            if self.node.model.existNode(outputNode):
                self.node.model.getNode(
                    outputNode).ioEngine.updateInputId(oldId, newId)

    def updateOnDeleteNode(self):
        for inputNode in self.inputs:
            if self.node.model.existNode(inputNode):
                self.node.model.getNode(inputNode).ioEngine.removeOutput(
                    self.node.identifier)

        for outputNode in self.outputs:
            if self.node.model.existNode(outputNode):
                self.node.model.getNode(outputNode).ioEngine.removeInput(
                    self.node.identifier)

    def updateOutputId(self, oldId, newId):
        if oldId in self.outputs:
            self.outputs.remove(oldId)
        self.outputs.append(newId)

    def updateInputId(self, oldId, newId):
        if oldId in self.inputs:
            self.inputs.remove(oldId)
        self.inputs.append(newId)

        self.node.updateDefinitionForChangeId(oldId, newId)
